"""
Benchmark for comparision to a strategy.
"""

import pinkfish.portfolio as portfolio
import pinkfish.pfstatistics as pfstatistics
import pinkfish.trade as trade
import pinkfish.utility as utility


class Benchmark:
    """
    Portfolio Benchmark for comparison to a strategy.
    """

    def __init__(self, symbols, capital, start, end,
                 dir_name='data',
                 use_adj=False,
                 use_continuous_calendar=False,
                 force_stock_market_calendar=False):
        """
        Initialize instance variables.

        Parameters
        ----------
        symbols : str or list of str
            The symbol(s) to use in the benchmark.
        capital : int
            The amount of money available for trading.
        start : datetime.datetime
            The desired start date for the benchmark.
        end : datetime.datetime
            The desired end date for the benchmark.
        dir_name : str, optional
            The leaf data dir name (default is 'data').
        use_adj : bool, optional
            True to adjust prices for dividends and splits
            (default is False).
        use_continuous_calendar: bool, optional
            True if your timeseries has data for all seven days a week,
            and you want to backtest trading every day, including weekends.
            If this value is True, then `force_stock_market_calendar`
            is set to False (default is False).
        force_stock_market_calendar : bool, optional
            True forces use of stock market calendar on timeseries.
            Normally, you don't need to do this.  This setting is intended
            to transform a continuous timeseries into a weekday timeseries.
            If this value is True, then `use_continuous_calendar` is set
            to False.

        Attributes
        ----------
        symbols : list of str
            The symbols to use in the benchmark.
        capital : int
            The amount of money available for trading.
        start : datetime.datetime
            The desired start date for the benchmark.
        end : datetime.datetime
            The desired end date for the benchmark.
        dir_name : str, optional
            The leaf data dir name (default is 'data').
        use_adj : bool, optional
            True to adjust prices for dividends and splits.
        use_continuous_calendar: bool, optional
            True if your timeseries has data for all seven days a week,
            and you want to backtest trading every day, including weekends.
            If this value is True, then `force_stock_market_calendar`
            is set to False (default is False).
        force_stock_market_calendar : bool, optional
            True forces use of stock market calendar on timeseries.
            Normally, you don't need to do this.  This setting is intended
            to transform a continuous timeseries into a weekday timeseries.
            If this value is True, then `use_continuous_calendar` is set
            to False.
        ts : pd.DataFrame
            The timeseries of the symbol used in backtest.
        rlog : pd.DataFrame
            The raw trade log.
        tlog : pd.DataFrame
            The trade log.
        dbal : pd.DataFrame
            The daily balance.
        stats : pd.Series
            The statistics for the benchmark.
        portfolio : pf.Portfolio
            The portfolio.
        """

        # If symbols is not a list, cast it to a list.
        if not isinstance(symbols, list):
            symbols = [symbols]

        symbols = list(set(symbols))

        self.symbols = symbols
        self.capital = capital
        self.start = start
        self.end = end
        self.dir_name = dir_name
        self.use_adj = use_adj
        self.use_continuous_calendar = use_continuous_calendar
        self.force_stock_market_calendar = force_stock_market_calendar

        self.ts = None
        self.rlog = None
        self.tlog = None
        self.dbal = None
        self.stats = None
        self.portfolio = None

    def _algo(self):
        """
        This is a general asset allocation stategy.

        Invest an equal percent in each investment option and
        rebalance every year.
        """
        trade.TradeLog.cash = self.capital
        trade.TradeLog.margin = trade.Margin.CASH

        # These dicts are used to track close and weights for
        # each symbol in portfolio
        weights = {}

        weight = 1 / len(self.portfolio.symbols)
        weights = {symbol:weight for symbol in self.portfolio.symbols}

        # Trading algorithm
        for i, row in enumerate(self.ts.itertuples()):

            end_flag = utility.is_last_row(self.ts, i)
            start_flag = (i==0)

            # Buy on first trading day
            # Rebalance on the first trading day of each year
            # Close all positions on last trading day
            if row.first_doty or end_flag or start_flag:

                # If last row, then zero out all weights.
                if end_flag:
                    weights = utility.set_dict_values(weights, 0)

                # Adjust weights of all symbols in portfolio
                self.portfolio.adjust_percents(row, weights, field='close')

            # Record daily balance.
            self.portfolio.record_daily_balance(row)

    def run(self):
        """
        Run the strategy.
        """
        self.portfolio = portfolio.Portfolio()
        self.ts = self.portfolio.fetch_timeseries(
            self.symbols, self.start, self.end,
            fields=['close'], dir_name=self.dir_name, use_adj=self.use_adj,
            use_continuous_calendar=self.use_continuous_calendar,
            force_stock_market_calendar=self.force_stock_market_calendar)
        # Add calendar columns
        self.ts = self.portfolio.calendar(self.ts)

        self.ts, self.start = self.portfolio.finalize_timeseries(self.ts, self.start)
        self.portfolio.init_trade_logs(self.ts)

        self._algo()
        self._get_logs()
        self._get_stats()

    def _get_logs(self):
        """
        Get the logs.
        """
        self.rlog, self.tlog, self.dbal = self.portfolio.get_logs()

    def _get_stats(self):
        """
        Get the stats.
        """
        self.stats = pfstatistics.stats(self.ts, self.tlog, self.dbal, self.capital)

Strategy = Benchmark
"""
class : Strategy is a class reference to Benchmark, i.e. an alias.
"""



"""
Plotting functions.
"""

import matplotlib.pyplot as plt
from pandas.plotting import register_matplotlib_converters
# Register matplotlib converters.
register_matplotlib_converters()

import pinkfish.pfstatistics as pfstatistics
import pinkfish.trade as trade


def plot_equity_curve(strategy, benchmark=None, yscale='linear', fname=None):
    """
    Plot Equity Curve: Strategy and (optionally) Benchmark.

    Parameters
    ----------
    strategy : pd.DataFrame
        Daily balance for the strategy.
    benchmark: pd.DataFrame, optional
        Daily balance for the benchmark (default is None, which implies
        that a benchmark is not being used).
    yscale: str, {'linear', 'log', 'symlog', 'logit'}
        The axis scale type to apply (default is 'linear')
    fname: str or path-like or file-like, optional
        Save the current figure to fname (default is None, which
        implies to not output the figure to a file).

    Returns
    -------
    None
    """
    fig = plt.figure()
    axes = fig.add_subplot(111, ylabel='Portfolio value in $')
    axes.plot(strategy['close'], label='strategy')
    axes.set_yscale(yscale)
    if benchmark is not None:
        axes.plot(benchmark['close'], label='benchmark')
    plt.legend(loc='best')
    if fname:
        plt.savefig(fname, bbox_inches='tight')

def plot_equity_curves(strategies, labels=None, yscale='linear', fname=None):
    """
    Plot one or more equity curves on the same plot.

    Parameters
    ----------
    strategies : pd.Series of pd.Dataframe
        Container of strategy daily balance for each symbol.
    labels : list of str, optional
        List of labels for each strategy (default is None, which implies
        that `strategy.symbol` is used as the label.
    yscale: str, {'linear', 'log', 'symlog', 'logit'}
        The axis scale type to apply (default is 'linear')
    fname: str or path-like or file-like, optional
        Save the current figure to fname (default is None, which
        implies to not output the figure to a file).

    Returns
    -------
    None
    """
    fig = plt.figure(figsize=(16,12))
    axes = fig.add_subplot(111, ylabel='Portfolio value in $')
    for i, strategy in enumerate(strategies):
        if labels is None:
            label = strategy.symbol
        else:
            label = labels[i]
        axes.plot(strategy.dbal['close'], label=label)
        axes.set_yscale(yscale)
    plt.legend(loc='best')
    if fname:
        plt.savefig(fname, bbox_inches='tight')

def plot_trades(strategy, benchmark=None, yscale='linear', fname=None):
    """
    Plot Trades.

    Benchmark is the equity curve that the trades get plotted on.
    If not provided, strategy equity curve is used.

    Parameters
    ----------
    strategy : pd.DataFrame
        Daily balance for the strategy.
    benchmark: pd.DataFrame, optional
        Daily balance for the benchmark.
    yscale: str, {'linear', 'log', 'symlog', 'logit'}
        The axis scale type to apply (default is 'linear')
    fname: str or path-like or file-like, optional
        Save the current figure to fname (default is None, which
        implies to not output the figure to a file).

    Returns
    -------
    None
    """
    if benchmark is None or strategy is benchmark:
        benchmark = strategy
        label = 'strategy'
    else:
        label = 'benchmark'

    fig = plt.figure()
    axes = fig.add_subplot(111, ylabel='Portfolio value in $')
    axes.plot(benchmark.index, benchmark['close'], label=label)
    axes.set_yscale(yscale)

    # Buy trades.
    s = strategy['state'] == trade.TradeState.OPEN
    s = s.reindex_like(benchmark)
    buy = benchmark[s]
    axes.plot(buy.index, buy['close'], '^', markersize=10, color='k')
    axes.set_yscale(yscale)

    # Sell trades.
    s = strategy['state'] == trade.TradeState.CLOSE
    s = s.reindex_like(benchmark)
    sell = benchmark[s]
    axes.plot(sell.index, sell['close'], 'v', markersize=10, color='r')
    axes.set_yscale(yscale)
    plt.legend(loc='best')
    if fname:
        plt.savefig(fname, bbox_inches='tight')


default_metrics = (
    'annual_return_rate',
    'max_closed_out_drawdown',
    'annualized_return_over_max_drawdown',
    'best_month',
    'worst_month',
    'sharpe_ratio',
    'sortino_ratio',
    'monthly_std',
    'annual_std')
"""
tuple : Default metrics for plot_bar_graph().

The metrics are:

    'annual_return_rate'  
    'max_closed_out_drawdown'  
    'annualized_return_over_max_drawdown'  
    'best_month'  
    'worst_month'  
    'sharpe_ratio'  
    'sortino_ratio'  
    'monthly_std'  
    'annual_std'
"""


def plot_bar_graph(stats, benchmark_stats=None, metrics=default_metrics,
                   extras=None, fname=None):
    """
    Plot Bar Graph: Strategy vs Benchmark (optional).

    Parameters
    ----------
    stats : pd.Series
        Statistics from the strategy.
    benchmark_stats : pd.Series, optional
        Statistics from the benchmark (default is None, which implies
        that a benchmark is not being used).
    metrics: tuple, optional
        The metrics to be plotted (default is `default_metrics`).
    extras: tuple, optional
        The additional metrics to be plotted (default is None, which
        implies no extra metrics should be added).
    fname: str or path-like or file-like, optional
        Save the current figure to fname (default is None, which
        implies to not output the figure to a file).

    Returns
    -------
    pd.DataFrame
        Summary metrics.
    """
    if extras is None:
        extras = ()
    metrics += extras

    df = pfstatistics.summary(stats, benchmark_stats, metrics)
    fig = plt.figure()
    axes = fig.add_subplot(111, ylabel='Trading Metrix')
    df.plot(kind='bar', ax=axes, color=['g', 'r'])
    axes.set_xticklabels(df.index, rotation=60)
    plt.legend(loc='best')
    if fname:
        plt.savefig(fname, bbox_inches='tight')
    return df


def optimizer_plot_bar_graph(df, metric):
    """
    Plot Bar Graph of a metric for a set of strategies.

    This function is designed to be used in analysis of an
    optimization of some parameter.  First call optimizer_summary()
    to generate the dataframe required by this function.

    Parameters
    ----------
    df : pd.DataFrame
        Summary of strategies vs metrics.
    metric : str
        The metric to be used in the summary.
    """
    df = df.loc[[metric]]
    df = df.transpose()
    fig = plt.figure()
    axes = fig.add_subplot(111, ylabel=metric)
    df.plot(kind='bar', ax=axes, legend=False)
    axes.set_xticklabels(df.index, rotation=0)



"""
Utility functions.
"""

from configparser import ConfigParser
from functools import wraps
import importlib.util
import inspect
import os
from pathlib import Path

import pandas as pd


ROOT = str(Path(os.getcwd().split('pinkfish')[0] + '/pinkfish'))
"""
str: pinkfish project root dir.
"""


def import_strategy(strategy_name, top_level_dir='examples', module_name='strategy'):
    """
    Import a strategy from a python `.py` file.

    Parameters
    ----------
    strategy_name : str
        The leaf dir name that contains the strategy to import.
    top_level_dir : str, optional
        The top level dir name for the strategies
        (default is 'examples').
    module_name: str, optional
        The name of the python module (default is 'strategy').

    Returns
    -------
    module
        The imported module.

    Examples
    --------
    >>> strategy = import_strategy(strategy_name='190.momentum-dmsr-portfolio')
    """

    strategy_location = (Path(ROOT) / Path(top_level_dir)
                                    / Path(strategy_name)
                                    / Path(module_name + '.py'))
    print(strategy_location)
    spec = importlib.util.spec_from_file_location(module_name, strategy_location)
    strategy = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(strategy)
    return strategy


def print_full(x):
    """
    Print every row of list-like object.
    """
    pd.set_option('display.max_rows', len(x))
    print(x)
    pd.reset_option('display.max_rows')


def read_config():
    """
    Read pinkfish configuration.
    """
    conf = {}
    parser = ConfigParser()
    parser.read(os.path.expanduser('~/.pinkfish'))
    conf['base_dir'] = parser.get('global', 'base_dir')
    return conf


def is_last_row(ts, index):
    """
    Return True for last row, False otherwise.
    """
    return True if (index == len(ts) - 1) else False


def sort_dict(d, reverse=False):
    """
    Return sorted dict; optionally reverse sort.
    """
    return dict(sorted(d.items(), key=lambda x: x[1], reverse=reverse))


def set_dict_values(d, value):
    """
    Return dict with same keys as `d` and all values equal to `value'.
    """
    return dict.fromkeys(d, value)


def find_nan_rows(ts):
    """
    Return a dataframe with the rows that contain NaN values.

    This function can help you track down problems with a timeseries.
    You may need to call `pd.set_option("display.max_columns", None)`
    at the top of your notebook to display all columns.

    Examples
    --------
    >>> pd.set_option("display.max_columns", None)
    >>> df = pf.find_nan_rows(ts)
    >>> df
    """
    is_NaN = ts.isnull()
    row_has_NaN = is_NaN.any(axis=1)
    rows_with_NaN = ts[row_has_NaN]
    return rows_with_NaN


def no_empty_container(container_name, default_ret_value):
    """
    Check if container is empty.  If so, return `default_ret_value`.

    Parameters
    ----------
    container_name : str
        The name of the container parameter to check.
    default_ret_value : int
        The return value the wrapped function if the container is empty.
        (default is 'examples').
    module_name: str, optional
        The name of the python module (default is 'strategy').

    Returns
    -------
    `default_ret_value` or `func` return value : type of return value
        If the container is empty, `default_ret_value` is returned,
        otherwise the return value of `func`.

    Examples
    --------
    >>> @no_empty_container('my_list', 0)
    >>> def my_func(my_list):
    >>>     return 5
    >>> my_func([])
    0
    """

    def decorator(func):
    
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                container_obj = kwargs[container_name]
            except KeyError:
                container_arg_index = inspect.getfullargspec(func).args.index(container_name)
                container_obj = args[container_arg_index]

            if len(container_obj) == 0:
                return default_ret_value

            return func(*args, **kwargs)
        return wrapper
    return decorator



"""
Analysis of results.

This module contains some functions that were copied or derived
from the book "Trading Evolved" by Andreas F. Clenow.
Below is a correspondance I had with the author:

------------------------------------------------------------------------
Farrell
October 25, 2019 at 15:49
Hi Andreas,

I just finished reading the book. Awesome one of a kind! Thanks so much.
I also enjoyed your other two. Question: what is the copyright (if any)
on the source code you have in the book. I want to incorporate some of
it into my open source backtester, Pinkfish. How should I credit your
work if no copyright.  I could add a comment at the beginning of each
derived function or module at a minimum.

Farrell
------------------------------------------------------------------------

Andreas Clenow
October 25, 2019 at 17:29
Hi Farrell,

I can be paid in reviews and/or beer. :)

For an open source project, use the code as you see fit. A credit in the
comments somewhere would be nice, but I won't sue you if you forget it.

ac
------------------------------------------------------------------------
"""

from IPython.core.display import display, HTML
import matplotlib.pyplot as plt
import pandas as pd

import empyrical as em

import pinkfish.indicator as indicator


########################################################################
# MONTHY RETURNS MAP

def monthly_returns_map(dbal):
    """
    Display per month and per year returns in a table.

    Parameters
    ----------
    dbal : pd.Series
        The daily closing balance indexed by date.

    Returns
    -------
    None

    Examples
    --------
    >>> monthly_returns_map(dbal['close'])
    Year    Jan     Feb     Mar     Apr     May     Jun     Jul ... Year
    1990   -8.5     0.9     2.4    -2.7     9.2    -0.9    -0.5    -8.2
    1991    4.2     6.7     2.2     0.0     3.9    -4.8     4.5    26.3
    """
    monthly_data = em.aggregate_returns(dbal.pct_change(), 'monthly')
    yearly_data = em.aggregate_returns(dbal.pct_change(), 'yearly')

    table_header = """
    <table class='table table-hover table-condensed table-striped'>
    <thead>
    <tr>
    <th style="text-align:right">Year</th>
    <th style="text-align:right">Jan</th>
    <th style="text-align:right">Feb</th>
    <th style="text-align:right">Mar</th>
    <th style="text-align:right">Apr</th>
    <th style="text-align:right">May</th>
    <th style="text-align:right">Jun</th>
    <th style="text-align:right">Jul</th>
    <th style="text-align:right">Aug</th>
    <th style="text-align:right">Sep</th>
    <th style="text-align:right">Oct</th>
    <th style="text-align:right">Nov</th>
    <th style="text-align:right">Dec</th>
    <th style="text-align:right">Year</th>
    </tr>
    </thead>
    <tbody>
    <tr>"""

    first_year = True
    first_month = True
    year = 0
    month = 0
    year_count = 0
    table = ''
    for m, val in monthly_data.items():
        year = m[0]
        month = m[1]

        if first_month:
            if year_count % 15 == 0:
                table += table_header
            table += f"<td align='right'><b>{year}</b></td>\n"
            first_month = False

        # Pad empty months for first year if sim doesn't start in Jan.
        if first_year:
            first_year = False
            if month > 1:
                for _ in range(1, month):
                    table += "<td align='right'>-</td>\n"

        table += "<td align='right'>{:.1f}</td>\n".format(val * 100)

        # Check for dec and add yearly.
        if month == 12:
            table += "<td align='right'><b>{:.1f}</b></td>\n".format(
                yearly_data[year] * 100)
            table += '</tr>\n <tr> \n'
            first_month = True
            year_count += 1

    # Add padding for empty months and last year's value.
    if month != 12:
        for i in range(month+1, 13):
            table += "<td align='right'>-</td>\n"
            if i == 12:
                table += "<td align='right'><b>{:.1f}</b></td>\n".format(
                    yearly_data[year] * 100)
                table += '</tr>\n <tr> \n'
    table += '</tr>\n </tbody> \n </table>'
    display(HTML(table))


########################################################################
# HOLDING PERIOD MAP

def holding_period_map(dbal):
    """
    Display holding period returns in a table.

    This shows what your annualized return would have been, had you
    started this strategy at the start of a given year, as shown in
    the leftmost column, and held it for a certain number of years.
    Length of returns should be 30 or less, otherwise the output
    will be jumbled.

    Parameters
    ----------
    dbal : pd.Series
        The daily closing balance indexed by date.

    Returns
    -------
    None

    Examples
    --------
    >>> table = holding_period_map(dbal['close'])
    >>> display(HTML(table))
    Years    1   2   3   4   5   6   7   8
    2013    30  20  13  12  13  10  12  12
    2014    11   5   7  10   6  10   9
    ...
    2020     8
    """
    year = em.aggregate_returns(dbal.pct_change(), 'yearly')
    year_start = 0

    table = "<table class='table table-hover table-condensed table-striped'>"
    table += "<tr><th>Years</th>"

    for i in range(len(year)):
        table += "<th>{}</th>".format(i+1)
    table += "</tr>"

    for the_year, _ in year.items(): # Iterates years
        table += f"<tr><th>{the_year}</th>"  # New table row

        for years_held in range(1, len(year)+1): # Iterates years held
            if years_held <= len(year.iloc[year_start:year_start + years_held]):
                ret = em.annual_return(year.iloc[year_start:year_start + years_held], 'yearly')
                table += "<td>{:.0f}</td>".format(ret * 100)
        table += "</tr>"
        year_start +=1
    display(HTML(table))


########################################################################
# PRETTIER GRAPHS

def _calc_corr(dbal, benchmark_dbal, window):
    """
    Calculate the rollowing correlation between two returns.

    Parameters
    ----------
    dbal : pd.Series
        Strategy daily closing balance indexed by date.
    benchmark_dbal : pd.Series
        Benchmark daily closing balance indexed by date.
    window : int
        Size of the moving window. This is the number of observations
        used for calculating the statistic.

    Returns
    -------
    corr : pd.DataFrame
        Window size rollowing correlation between `dbal` and
        `benchmark_dbal`.
    """
    ret = dbal.pct_change()
    benchmark_ret = benchmark_dbal.pct_change()
    corr = ret.rolling(window).corr(benchmark_ret)
    return corr


def prettier_graphs(dbal, benchmark_dbal, dbal_label='Strategy',
                    benchmark_label='Benchmark', points_to_plot=None):
    """
    Plot 3 subplots.

    The first subplot will show a rebased comparison of the returns to
    the benchmark returns, recalculated with the same starting value
    of 1.  This will be shown on a semi logarithmic scale.  The second
    subplot will show relative strength of the returns to the benchmark
    returns, and the third the correlation between the two.

    Parameters
    ----------
    dbal : pd.Series
        Strategy daily closing balance indexed by date.
    benchmark_dbal : pd.Series
        Benchmark daily closing balance indexed by date.
    label : str, optional
        Label to use in graph for strategy (default is 'Strategy').
    benchmark_label : str, optional
        Label to use in graph for benchmark (default is 'Benchmark').
    points_to_plot : int, optional
        Define how many points (trading days) we intend to plot
        (default is None, which implies plot all points or days).

    Returns
    -------
    None

    Examples
    --------
    >>> prettier_graphs(dbal['close'], benchmark_dbal['close'],
                        points_to_plot=5000)
    """
    if points_to_plot is None:
        points_to_plot = 0

    data = pd.DataFrame(dbal)
    data['benchmark_dbal'] = pd.DataFrame(benchmark_dbal)
    data.columns = ['dbal', 'benchmark_dbal']
    data.head()

    # Rebase the two series to the same point in time;
    # starting where the plot will start.
    for col in data:
        data[col + '_rebased'] = \
            (data[-points_to_plot:][col].pct_change() + 1).cumprod()

    # Relative strength, strategy to benchmark.
    data['relative_strength'] = data['dbal'] / data['benchmark_dbal']

    # Calculate 100 day rolling correlation.
    data['corr'] = _calc_corr(data['dbal'], data['benchmark_dbal'], 100)

    # After this, we slice the data, effectively discarding all but
    # the last points_to_plot data points, using the slicing logic from
    # before.  Slice the data, cut points we don't intend to plot.
    plot_data = data[-points_to_plot:]

    # Make new figure and set the size.
    fig = plt.figure(figsize=(12, 8))

    # The first subplot, planning for 3 plots high, 1 plot wide,
    # this being the first.
    ax = fig.add_subplot(311)
    ax.set_title('Comparison')
    ax.semilogy(plot_data['dbal_rebased'], linestyle='-',
                label=dbal_label, linewidth=3.0)
    ax.semilogy(plot_data['benchmark_dbal_rebased'], linestyle='--',
                label=benchmark_label, linewidth=3.0)
    ax.legend()
    ax.grid(False)

    # Second sub plot.
    ax = fig.add_subplot(312)
    label = f'Relative Strength, {dbal_label} to {benchmark_label}'
    ax.plot(plot_data['relative_strength'], label=label, linestyle=':', linewidth=3.0)
    ax.legend()
    ax.grid(True)

    # Third subplot.
    ax = fig.add_subplot(313)
    label = f'Correlation between {dbal_label} and {benchmark_label}'
    ax.plot(plot_data['corr'], label=label, linestyle='-.', linewidth=3.0)
    ax.legend()
    ax.grid(True)


########################################################################
# VOLATILITY

def volatility_graphs(dbals, labels, points_to_plot=None):
    """
    Plot volatility graphs.

    The first graph is a boxplot showing the differences between
    2 or more returns.  The second graph shows the volatility plotted
    for 2 or more returns.

    Parameters
    ----------
    dbals : list of pd.DataFrame
        A list of daily closing balances (or daily instrument closing
        prices) indexed by date.
    labels : list of str
        A list of labels.
    points_to_plot : int, optional
        Define how many points (trading days) we intend to plot
        (default is None, which implies plot all points or days).

    Returns
    -------
    pf.DataFrame
        Statistics comparing the `dbals`.

    Examples
    --------
    >>> df = pf.volatility_graph([ts, dbal], ['SPY', 'Strategy'],
                                 points_to_plot=5000)
    >>> df
    """
    def _boxplot(volas, labels):
        """
        Plot a volatility boxplot.
        """
        fig = plt.figure(figsize=(12, 8))
        fig.add_subplot(111, ylabel='Volatility')
        plt.ylim(0, 1)
        plt.boxplot(volas, labels=labels)


    def _volas_plot(volas, labels):
        """
        Plot volatility.
        """
        fig = plt.figure(figsize=(14, 10))
        axes = fig.add_subplot(111, ylabel='Volatility')
        for i, vola in enumerate(volas):
            axes.plot(vola, label=labels[i])
        plt.legend(loc='best')

    if points_to_plot is None:
        points_to_plot = 0

    # Get volatility for each dbal set.
    volas = []
    for dbal in dbals:
        volas.append(indicator.VOLATILITY(dbal[-points_to_plot:]).dropna())

    # Build metrics dataframe.
    index = []
    columns = labels
    data = []
    # Add metrics.
    metrics = ['avg', 'median', 'min', 'max', 'std', 'last']
    for metric in metrics:
        index.append(metric)
        if   metric == 'avg':    data.append(vola.mean() for vola in volas)
        elif metric == 'median': data.append(vola.median() for vola in volas)
        elif metric == 'min':    data.append(vola.min() for vola in volas)
        elif metric == 'max':    data.append(vola.max() for vola in volas)
        elif metric == 'std':    data.append(vola.std() for vola in volas)
        elif metric == 'last':   data.append(vola.iloc[-1] for vola in volas)

    df = pd.DataFrame(data, columns=columns, index=index)
    _boxplot(volas, labels)
    _volas_plot(volas, labels)
    return df


########################################################################
# KELLY CRITERION

def kelly_criterion(stats, benchmark_stats=None):
    """
    Use this function to help with sizing of leverage.

    This function uses ideas based on the Kelly Criterion.

    Parameters
    ----------
    stats : pd.Series
        Statistics for the strategy.
    bbenchmark_stats : pd.Series, optimal
        Statistics for the benchmark (default is None, which implies
        that a benchmark is not being used).

    Returns
    -------
    s : pf.Series
        Leverage statistics.

         - `sharpe_ratio` is a measure of risk adjusted return.

         - `sharpe_ratio_max` is the maximum expected sharpe ratio.

         - `sharpe_ratio_min` is the minimum expected sharpe ratio.

         - `strategy risk` is a measure of how risky a trading strategy
            is, calculated as an annual standard deviation of returns.

         - `instrument_risk` is a measure of how risky an instrument is
            before any leverage is applied, calculated as an annual
            standard deviation of returns.

         - `optimal target risk` is equal to the expected sharpe ratio,
            according to the Kelly criterion.  Target risk is the amount
            of risk you expect to see when trading, calculated as an
            annual standard deviation of returns.

         - `half kelly criterion` is equal to half the expected
            sharpe ratio.  It uses a conservative version of the
            Kelly criterion known as half Kelly.

         - `aggressive leverage` is the optimal target risk divided by
            the instrument risk.  This is an aggressive form of the
            leverage factor, which is the cash value of a position
            divided by your capital.

         - `moderate leverage` is the leverage factor calculated using
            half Kelly.

         - `conservative leverage` is the leverage factor calculated
            using half of the minimum sharpe ratio divided by 2.
    """
    s = pd.Series(dtype='object')

    s['sharpe_ratio']           = stats['sharpe_ratio']
    s['sharpe_ratio_max']       = stats['sharpe_ratio_max']
    s['sharpe_ratio_min']       = stats['sharpe_ratio_min']
    s['strategy risk']          = stats['annual_std'] / 100
    if benchmark_stats is not None:
        s['instrument risk']    = benchmark_stats['annual_std'] / 100
    s['optimal target risk']    = s['sharpe_ratio']
    s['half kelly criterion']   = s['sharpe_ratio'] / 2
    s['aggressive leverage']    = s['optimal target risk'] / s['instrument risk']
    s['moderate leverage']      = s['half kelly criterion'] / s['instrument risk']
    s['conservative leverage']  = (s['sharpe_ratio_min'] / 2) / s['instrument risk']
    return s



from .fetch import (
    fetch_timeseries,
    select_tradeperiod,
    finalize_timeseries,
    remove_cache_symbols,
    update_cache_symbols,
    get_symbol_metadata
)

from .trade import (
    Direction,
    Margin,
    TradeLog,
    TradeState,
    DailyBal
)

from .pfstatistics import (
    ALPHA_BEGIN,
    SP500_BEGIN,
    get_trading_days,
    currency_metrics,
    stats,
    currency,
    summary,
    optimizer_summary
)

from .plot import (
    plot_equity_curve,
    plot_equity_curves,
    plot_trades,
    plot_bar_graph,
    optimizer_plot_bar_graph
)

from .benchmark import (
    Benchmark
)

from .portfolio import (
    Portfolio,
    technical_indicator
)

from .indicator import (
    SMA,
    EMA,
    CROSSOVER,
    MOMENTUM,
    VOLATILITY,
    ANNUALIZED_RETURNS,
    ANNUALIZED_STANDARD_DEVIATION,
    ANNUALIZED_SHARPE_RATIO
)

from .pfcalendar import (
    calendar
)

from .stock_market_calendar import (
    stock_market_calendar
)

from .analysis import (
    monthly_returns_map,
    holding_period_map,
    prettier_graphs,
    volatility_graphs,
    kelly_criterion
)

from .utility import (
    ROOT,
    import_strategy,
    print_full,
    read_config,
    is_last_row,
    sort_dict,
    set_dict_values,
    find_nan_rows
)

DEBUG = False
"""
bool : True to enable DBG() output.
"""
def DBG(s):
    """
    Debug print.  Enable by setting pf.DEBUG=True.
    """
    if DEBUG:
        print(s)
    else:
        pass



"""
Custom indicators.

These indicators are meant to supplement the TA-Lib.  See:
https://ta-lib.org/function.html
"""

import math
import numpy as np
import pandas as pd

import pinkfish.pfstatistics as pfstatistics


class IndicatorError(Exception):
    """
    Base indicator exception.
    """
    pass


########################################################################
# SMA

def SMA(ts, timeperiod=30, price='close'):
    """
    This indicator computes a simple moving average.

    Can be used in place of talib SMA.

    ts : pd.DateFrame or pd.Series
        A dataframe with 'open', 'high', 'low', 'close', 'volume' or
        a series of price data.
    timeperiod: int, optional
        The timeperiod for the moving average (default is 30).
    price : str, optional {'close', 'open', 'high', 'low'}
        Input_array column to use for price (default is 'close').
        Not used if `ts` is a series.

    Returns
    -------
    pd.Series
        Series that contains the simple moving average.

    Examples
    --------
    >>> ts['sma50'] = pf.SMA(ts, timeperiod=50)
    """
    s = ts[price] if isinstance(ts, pd.DataFrame) else ts
    return s.rolling(timeperiod).mean()


########################################################################
# EMA

def EMA(ts, timeperiod=30, price='close'):
    """
    This indicator computes an exponential moving average.

    Can be used in place of talib EMA.

    ts : pd.DateFrame or pd.Series
        A dataframe with 'open', 'high', 'low', 'close', 'volume' or
        a series of price data.
    timeperiod: int, optional
        The timeperiod for the moving average (default is 30).
    price : str, optional {'close', 'open', 'high', 'low'}
        Input_array column to use for price (default is 'close').
        Not used if `ts` is a series.

    Returns
    -------
    pd.Series
        Series that contains the simple moving average.

    Examples
    --------
    >>> ts['ema50'] = pf.EMA(ts, timeperiod=50)
    """
    s = ts[price] if isinstance(ts, pd.DataFrame) else ts
    return s.ewm(span=timeperiod, min_periods=timeperiod, adjust=False).mean()


########################################################################
# CROSSOVER

class TradeCrossOverError(IndicatorError):
    """
    Invalid timeperiod specified.
    """
    pass


class _CrossOver:
    """
    This is a helper class to implement the CROSSOVER function.

     The class provides the apply callback for pd.DataFrame.apply()
     in CROSSOVER.  It also keeps track of _r, explained below.

    _r indicates regime direction and duration, i.e. 50 means a bull
       market that has persisted for 50 days, whereas -20 means a bear
       market that has persisted for 20 days.
    _r is incremented(decremented) each day a bull(bear) market persists
    _r remains unchanged when fast_ma within band of slow_ma
    _r indicates the number of trading days a trend has persisted
    _r is nan, then sma_slow is nan
    _r > 0, then bull market, fast_ma > slow_ma
    _r < 0, then bear market, fast_ma < slow_ma
    _r == 0, no trend established yet
    """
    def __init__(self):
        """
        Initialize instance variables.

        Attributes
        ----------
        _r : int
            Indicates regime direction and duration.
        """
        self._r = 0

    def apply(self, row, band=0):
        """
        Implements the regime change logic.

        Parameters
        ----------
        row : pd.Series
            A row of data from the dataframe.
        band : int {0-100}
            Percent band (default is 0, which is no band).

        Returns
        -------
        _r : int
            Indicates regime direction and duration.
        """
        if pd.isnull(row['__sma_slow__']):
            self._r = np.nan
        elif row['__sma_fast__'] > row['__sma_slow__']*(1+band/100):
            self._r = self._r + 1 if self._r > 0 else 1
        elif row['__sma_fast__'] < row['__sma_slow__']*(1-band/100):
            self._r = self._r -1 if self._r < 0 else -1
        else:
            pass
        return self._r


def CROSSOVER(ts, timeperiod_fast=50, timeperiod_slow=200,
              func_fast=SMA, func_slow=SMA, band=0,
              price='close', prevday=False):
    """
    This indicator is used to represent regime direction and duration.

    For example, an indicator value of 50 means a bull market that has
    persisted for 50 days, whereas -20 means a bear market that has
    persisted for 20 days.

    More generally, this is a crossover indicator for two moving
    averages.  The indicator is positive when the fast moving average
    is above the slow moving arverage, and negative when the fast
    moving average is below the slow moving average.

    Parameters
    ----------
    ts : pd.DateFrame
        A dataframe with 'open', 'high', 'low', 'close', 'volume'.
    timeperiod_fast : int, optional
        The timeperiod for the fast moving average (default is 50).
    timeperiod_slow : int, optional
        The timeperiod for the slow moving average (default is 200).
    func_fast : Function, optional
        {pf.SMA, pf.EMA} (pinkfish functions) or
        {SMA, DEMA, EMA, KAMA, T3, TEMA, TRIMA, WMA} (ta-lib functions)
        The function for fast moving average (default is pf.SMA).
        MAMA not compatible.
    func_slow : Function, optional
         {pf.SMA, pf.EMA} (pinkfish functions) or
        {SMA, DEMA, EMA, KAMA, T3, TEMA, TRIMA, WMA} (ta-lib functions)
        The function for fast moving average (default is pf.SMA).
        MAMA not compatible.
    band : float, {0-100}, optional
        Percent band around the slow moving average.
        (default is 0, which implies no band is used).
    price : str, optional {'close', 'open', 'high', 'low'}
        Input_array column to use for price (default is 'close').
    prevday : bool, optional
        True will shift the series forward.  Unless you are buying
        on the close, you'll likely want to set this to True.
        It gives you the previous day's CrossOver (default is False).

    Returns
    -------
    s : pd.Series
        Series that contains the rolling regime indicator values.

    Raises
    ------
    TradeCrossOverError
        If one of the timeperiods specified is invalid.

    Examples
    --------
    >>> ts['regime'] = pf.CROSSOVER(ts, timeperiod_fast=50,
                                    timeperiod_slow=200)
    """
    if (timeperiod_fast < 1 or timeperiod_slow < 2
        or timeperiod_fast >= timeperiod_slow):
        raise TradeCrossOverError

    ts['__sma_fast__'] = ts[price] if timeperiod_fast == 1 else \
        func_fast(ts, timeperiod=timeperiod_fast, price=price)

    ts['__sma_slow__'] = \
        func_slow(ts, timeperiod=timeperiod_slow, price=price)

    func = _CrossOver().apply
    s = ts.apply(func, band=band, axis=1)
    if prevday:
        s = s.shift()
    ts.drop(['__sma_fast__', '__sma_slow__'], axis=1, inplace=True)
    return s


########################################################################
# MOMENTUM

def MOMENTUM(ts, lookback=1, time_frame='monthly', price='close', prevday=False):
    """
    This indicator is used to represent momentum is security prices.

    Percent price change is used to calculate momentum.  Momentum
    is positive if the price since the lookback period has increased.
    Likewise, if price has decreased since the lookback period,
    momentum is negative.  Percent change is used to normalize
    asset prices for comparison.

    Parameters
    ----------
    ts : pd.DateFrame
        A dataframe with 'open', 'high', 'low', 'close', 'volume'.
    lookback : int, optional
        The number of time frames to lookback, e.g. 2 months
        (default is 1).
    timeframe : str, optional {'monthly', 'daily', 'weekly', 'yearly'}
        The unit or timeframe type of lookback (default is 'monthly').
    price : str, optional {'close', 'open', 'high', 'low'}
        Input_array column to use for price (default is 'close').
    prevday : bool, optional
        True will shift the series forward.  Unless you are buying
        on the close, you'll likely want to set this to True.
        It gives you the previous day's Momentum (default is False).

    Returns
    -------
    s : pd.Series
        Series that contains the rolling momentum indicator values.

    Raises
    ------
    ValueError
        If the lookback is not positive or the time_frame is invalid.

    Examples
    --------
    >>> ts['mom'] = pf.MOMENTUM(ts, lookback=6, time_frame='monthly')
    """
    if lookback < 1:
        raise ValueError('lookback must be positive')

    if   time_frame =='daily':   factor = 1
    elif time_frame =='weekly':  factor = pfstatistics.TRADING_DAYS_PER_WEEK
    elif time_frame =='monthly': factor = pfstatistics.TRADING_DAYS_PER_MONTH
    elif time_frame =='yearly':  factor = pfstatistics.TRADING_DAYS_PER_YEAR
    else:
        raise ValueError(f'invalid time_frame "{time_frame}"')

    s = ts[price].pct_change(periods=lookback*factor)
    if prevday:
        s = s.shift()

    return s


########################################################################
# VOLATILITY

def VOLATILITY(ts, lookback=20, time_frame='yearly', downside=False,
               price='close', prevday=False):
    """
    This indicator is used to represent volatility in security prices.

    Volatility is represented as the standard deviation.  Volatility
    is calculated over the lookback period, then we scale to the
    time frame.  Volatility scales with the square root of time.
    For example,  if the market’s daily volatility is 0.5%, then
    volatility for two days is the square root of 2 times
    the daily volatility (0.5% * 1.414 = 0.707%).  We use the square
    root of time to scale from daily to weely, monthly, or yearly.

    Parameters
    ----------
    ts : pd.DateFrame
        A dataframe with 'open', 'high', 'low', 'close', 'volume'.
    lookback : int, optional
        The number of time frames to lookback, e.g. 2 months
        (default is 1).
    timeframe : str, optional {'yearly', 'daily', 'weekly', 'monthly'}
        The unit or timeframe used for scaling.  For example, if the
        lookback is 20 and the timeframe is 'yearly', then we compute
        the 20 day volatility and scale to 1 year.
        (default is 'yearly').
    downside : bool, optional
        True to calculate the downside volatility (default is False).
    price : str, optional {'close', 'open', 'high', 'low'}
        Input_array column to use for price (default is 'close').
    prevday : bool, optional
        True will shift the series forward.  Unless you are buying
        on the close, you'll likely want to set this to True.
        It gives you the previous day's Volatility (default is False).

    Returns
    -------
    s : pd.Series
        A new column that contains the rolling volatility.

    Raises
    ------
    ValueError
        If the lookback is not positive or the time_frame is invalid.

    Examples
    --------
    >>> ts['vola'] = pf.VOLATILITY(ts, lookback=20, time_frame='yearly')
    """
    if lookback < 1:
        raise ValueError('lookback must be positive')

    if   time_frame == 'daily':   factor = 1
    elif time_frame == 'weekly':  factor = pfstatistics.TRADING_DAYS_PER_WEEK
    elif time_frame == 'monthly': factor = pfstatistics.TRADING_DAYS_PER_MONTH
    elif time_frame == 'yearly':  factor = pfstatistics.TRADING_DAYS_PER_YEAR
    else:
        raise ValueError(f'invalid time_frame "{time_frame}"')

    s = ts[price].pct_change()
    if downside:
        s[s > 0] = 0
    s = s.rolling(window=lookback).std() * np.sqrt(factor)

    if prevday:
        s = s.shift()

    return s


########################################################################
# ANNUALIZED_RETURNS

def ANNUALIZED_RETURNS(ts, lookback=5, price='close', prevday=False):
    """
    Calculate the rolling annualized returns.

    Parameters
    ----------
    ts : pd.DateFrame
        A dataframe with 'open', 'high', 'low', 'close', 'volume'.
    lookback : float, optional
        The number of years to lookback, e.g. 5 years.  1/12 can be
        used for 1 month.  Likewise 3/12 for 3 months, etc...
        (default is 5).
    price : str, optional {'close', 'open', 'high', 'low'}
        Input_array column to use for price (default is 'close').
    prevday : bool, optional
        True will shift the series forward.  Unless you are buying
        on the close, you'll likely want to set this to True.
        It gives you the previous day's Volatility (default is False).

    Returns
    -------
    s : pd.Series
        Series that contains the rolling annualized returns.

    Raises
    ------
    ValueError
        If the lookback is not positive.

    Examples
    --------
    >>> annual_returns_1mo = pf.ANNUALIZED_RETURNS(ts, lookback=1/12)
    >>> annual_returns_3mo = pf.ANNUALIZED_RETURNS(ts, lookback=3/12)
    >>> annual_returns_1yr = pf.ANNUALIZED_RETURNS(ts, lookback=1)
    >>> annual_returns_3yr = pf.ANNUALIZED_RETURNS(ts, lookback=3)
    >>> annual_returns_5yr = pf.ANNUALIZED_RETURNS(ts, lookback=5)
    """
    def _cagr(s):
        """
        Calculate compound annual growth rate.

        B = end balance; A = begin balance; n = num years
        """
        A = s.iloc[0]
        B = s.iloc[-1]
        n = len(s)
        if B < 0: B = 0
        return (math.pow(B / A, 1 / n) - 1) * 100

    if lookback <= 0:
        raise ValueError('lookback must be positive')

    window = int(lookback * pfstatistics.TRADING_DAYS_PER_YEAR)
    s = pd.Series(ts[price]).rolling(window).apply(_cagr)
    if prevday:
        s = s.shift()

    return s


########################################################################
# ANNUALIZED_STANDARD_DEVIATION

def ANNUALIZED_STANDARD_DEVIATION(ts, lookback=3, price='close', prevday=False):
    """
    Calculate the rolling annualized standard deviation.

    Parameters
    ----------
    ts : pd.DateFrame
        A dataframe with 'open', 'high', 'low', 'close', 'volume'.
    lookback : float, optional
        The number of years to lookback, e.g. 5 years.  1/12 can be
        used for 1 month.  Likewise 3/12 for 3 months, etc...
        (default is 5).
    price : str, optional {'close', 'open', 'high', 'low'}
        Input_array column to use for price (default is 'close').
    prevday : bool, optional
        True will shift the series forward.  Unless you are buying
        on the close, you'll likely want to set this to True.
        It gives you the previous day's Volatility (default is False).

    Returns
    -------
    s : pd.Series
        Series that contains the rolling annualized standard deviation.

    Raises
    ------
    ValueError
        If the lookback is not positive.

    Examples
    --------
    >>> std_dev_1mo = pf.ANNUALIZED_STANDARD_DEVIATION(ts,lookback=1/12)
    >>> std_dev_3mo = pf.ANNUALIZED_STANDARD_DEVIATION(ts, lookback=3/12)
    >>> std_dev_1yr = pf.ANNUALIZED_STANDARD_DEVIATION(ts, lookback=1)
    >>> std_dev_3yr = pf.ANNUALIZED_STANDARD_DEVIATION(ts, lookback=3)
    >>> std_dev_5yr = pf.ANNUALIZED_STANDARD_DEVIATION(ts, lookback=5)
    """
    def _std_dev(s):
        """
        Calculate the annualized standard deviation.
        """
        return np.std(s, axis=0) * math.sqrt(pfstatistics.TRADING_DAYS_PER_YEAR)

    if lookback <= 0:
        raise ValueError('lookback must be positive')

    window = int(lookback * pfstatistics.TRADING_DAYS_PER_YEAR)
    pc = ts[price].pct_change()
    s = pd.Series(pc).rolling(window).apply(_std_dev)
    if prevday:
        s = s.shift()

    return s


########################################################################
# ANNUALIZED_SHARPE_RATIO

def ANNUALIZED_SHARPE_RATIO(ts, lookback=5, price='close', prevday=False,
                            risk_free=0):
    """
    Calculate the rolling annualized sharpe ratio.

    Parameters
    ----------
    ts : pd.DateFrame
        A dataframe with 'open', 'high', 'low', 'close', 'volume'.
    lookback : float, optional
        The number of years to lookback, e.g. 5 years.  1/12 can be
        used for 1 month.  Likewise 3/12 for 3 months, etc...
        (default is 5).
    price : str, optional {'close', 'open', 'high', 'low'}
        Input_array column to use for price (default is 'close').
    prevday : bool, optional
        True will shift the series forward.  Unless you are buying
        on the close, you'll likely want to set this to True.
        It gives you the previous day's Volatility (default is False).
    risk_free: float, optional
        The risk free rate (default is 0).

    Returns
    -------
    s : pd.Series
        Series that contains the rolling annualized sharpe ratio.

    Raises
    ------
    ValueError
        If the lookback is not positive.

    Examples
    --------
    >>> sharpe_ratio_1mo = pf.ANNUALIZED_SHARPE_RATIO(ts, lookback=1/12)
    >>> sharpe_ratio_3mo = pf.ANNUALIZED_SHARPE_RATIO(ts, lookback=3/12)
    >>> sharpe_ratio_1yr = pf.ANNUALIZED_SHARPE_RATIO(ts, lookback=1)
    >>> sharpe_ratio_3yr = pf.ANNUALIZED_SHARPE_RATIO(ts, lookback=3)
    >>> sharpe_ratio_5yr = pf.ANNUALIZED_SHARPE_RATIO(ts, lookback=5)
    """
    def _sharpe_ratio(s):
        """
        Calculate the annualized sharpe ratio.
        """
        dev = np.std(s, axis=0)
        mean = np.mean(s, axis=0)
        period = len(s)
        sharpe = (mean*period - risk_free) / (dev * np.sqrt(period))
        return sharpe

    if lookback <= 0:
        raise ValueError('lookback must be positive')

    window = int(lookback * pfstatistics.TRADING_DAYS_PER_YEAR)
    pc = ts[price].pct_change()
    s = pd.Series(pc).rolling(window).apply(_sharpe_ratio)
    if prevday:
        s = s.shift()

    return s



"""
Portfolio backtesting.
"""

from functools import wraps

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn

from pinkfish.pfcalendar import calendar
from pinkfish.fetch import (
    fetch_timeseries,
    select_tradeperiod,
    finalize_timeseries
)
import pinkfish.pfstatistics as pfstatistics
import pinkfish.trade as trade
import pinkfish.utility as utility


def technical_indicator(symbols, output_column_suffix,
                        input_column_suffix='close'):
    """
    Decorator for adding a technical indicator to portfolio symbols.

    A new column will be added for each symbol.  The name of the
    new column will be the symbol name, an underscore, and the
    `output_column_suffix`.  For example, 'SPY_MA30' is the symbol
    SPY with `output_column_suffix` equal to MA30.

    `func` is a wrapper for a technical analysis function.  The
    actual technical analysis function could be from ta-lib,
    pandas, pinkfish indicator, or a custom user function.

    'func' must have the positional argument `ts` and keyword argument
    `input_column`.  'ts` is passed in, but input_column (args[1]) is
    assigned in the wrapper before `func` is called.

    Parameters
    ----------
    symbols : list
        The symbols that constitute the portfolio.
    output_column_suffix : str
        Output column suffix to use for technical indicator.
    input_column_suffix : str, {'close', 'open', 'high', 'low'}
        Input column suffix to use for price (default is 'close').

    Returns
    -------
    decorator : function
        A wrapper that adds technical indicators to portfolio
        symbols.

    Examples
    --------
    >>> # Technical indicator: volatility.
    >>> @pf.technical_indicator(symbols, 'vola', 'close')
    >>> def _volatility(ts, input_column=None):
    ...     return pf.VOLATILITY(ts, price=input_column)
    >>> ts = _volatility(ts)
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            assert len(args) >= 1, f'func requires at least 1 args, detected {len(args)}'
            assert type(args[0]) == pd.DataFrame, f'args[0] not a pd.DataFrame'
            ts = args[0]
            indicator_column = {}
            for symbol in symbols:
                input_column = symbol + '_' + input_column_suffix
                output_column = symbol + '_' + output_column_suffix
                kwargs['input_column'] = input_column
                indicator_column[output_column] = func(*args, **kwargs)
            
            # Join all the symbol columns to the original DataFrame using pd.concat
            ts = pd.concat([ts, pd.DataFrame(indicator_column)], axis=1)
            return ts
        return wrapper
    return decorator


class Portfolio:
    """
    A portfolio or collection of securities.

    Methods
    -------
     - fetch_timeseries()
       Get time series data for symbols.

     - add_technical_indicator()
       Add a technical indicator for each symbol in the portfolio.

     - calendar()
       Add calendar columns.

     - finalize_timeseries()
       Finalize timeseries.

     - get_price()
       Return price given row, symbol, and field.

     - get_prices()
       Return dict of prices for all symbols given row and fields.

     - shares()
       Return number of shares for given symbol in portfolio.

     - positions
       Gets the active symbols in portfolio as a list.

     - share_percent()
       Return share value of symbol as a percentage of `total_funds`.

     - adjust_percent()
       Adjust symbol to a specified weight (percent) of portfolio.

     - print_holdings()
       Print snapshot of portfolio holding and values.

     - init_trade_logs()
       Add a trade log for each symbol.

     - record_daily_balance()
       Append to daily balance list.

     - get_logs()
       Return raw tradelog, tradelog, and daily balance log.

     - performance_per_symbol()
       Returns performance per symbol data, also plots performance.

     - correlation_map()
       Show correlation map between symbols.
    """

    def __init__(self):
        """
        Initialize instance variables.

        Attributes
        ----------
        _l : list of tuples
            The list of daily balance tuples.
        _ts : pd.DataFrame
            The timeseries of the portfolio.
        symbols : list
            The symbols that constitute the portfolio.
        """
        self._l = []
        self._ts = None
        self.symbols = []

    ####################################################################
    # TIMESERIES (fetch, add_technical_indicator, calender, finalize)

    def _add_symbol_columns(self, ts, symbol, symbol_ts, fields):
        """
        Add column with field suffix for symbol, i.e. SPY_close.
        """
        for field in fields:
            column = symbol + '_' + field
            ts[column] = symbol_ts[field]
        return ts

    def fetch_timeseries(self, symbols, start, end,
                         fields=['open', 'high', 'low', 'close'],
                         dir_name='data',
                         use_cache=True, use_adj=True,
                         use_continuous_calendar=False,
                         force_stock_market_calendar=False,
                         check_fields=['close']):
        """
        Fetch time series data for symbols.

        Parameters
        ----------
        symbols : list
            The list of symbols to fetch timeseries.
        start : datetime.datetime
            The desired start date for the strategy.
        end : datetime.datetime
            The desired end date for the strategy.
        fields : list, optional
            The list of fields to use for each symbol (default is
            ['open', 'high', 'low', 'close']).  List must include 
            'close' - will be added if not already in list.
        dir_name : str, optional
            The leaf data dir name (default is 'data').
        use_cache: bool, optional
            True to use data cache.  False to retrieve from the
            internet (default is True).
        use_adj : bool, optional
            True to adjust prices for dividends and splits
            (default is False).
        use_continuous_calendar: bool, optional
            True if your timeseries has data for all seven days a week,
            and you want to backtest trading every day, including weekends.
            If this value is True, then `force_stock_market_calendar`
            is set to False (default is False).
        force_stock_market_calendar : bool, optional
            True forces use of stock market calendar on timeseries.
            Normally, you don't need to do this.  This setting is intended
            to transform a continuous timeseries into a weekday timeseries.
            If this value is True, then `use_continuous_calendar` is set
            to False (default is False).
        check_fields : list of str, optional {'high', 'low', 'open',
            'close', 'adj_close'}
            Fields to check for for NaN values.  If a NaN value is found
            for one of these fields, that row is dropped
            (default is ['close']).

        Returns
        -------
        pd.DataFrame
            The timeseries of the symbols.
        """
        if 'close' not in fields:
            fields.append('close')

        symbols = list(set(symbols))
        for i, symbol in enumerate(symbols):

            if i == 0:
                ts = fetch_timeseries(symbol, dir_name=dir_name, use_cache=use_cache)
                ts = select_tradeperiod(ts, start, end, use_adj=use_adj,
                                        use_continuous_calendar=use_continuous_calendar,
                                        force_stock_market_calendar=force_stock_market_calendar,
                                        check_fields=check_fields)
                self._add_symbol_columns(ts, symbol, ts, fields)
                ts.drop(columns=['open', 'high', 'low', 'close', 'volume', 'adj_close'],
                        inplace=True)
            else:
                # Add another symbol.
                _ts = fetch_timeseries(symbol, dir_name=dir_name, use_cache=use_cache)
                _ts = select_tradeperiod(_ts, start, end, use_adj=use_adj,
                                         use_continuous_calendar=use_continuous_calendar,
                                         force_stock_market_calendar=force_stock_market_calendar,
                                         check_fields=check_fields)
                self._add_symbol_columns(ts, symbol, _ts, fields)

        ts.dropna(inplace=True)
        self.symbols = symbols
        return ts

    def add_technical_indicator(self, ts, ta_func, ta_param, output_column_suffix,
                                input_column_suffix='close'):
        """
        Add a technical indicator for each symbol in the portfolio.

        A new column will be added for each symbol.  The name of the
        new column will be the symbol name, an underscore, and the
        `output_column_suffix`.  For example, 'SPY_MA30' is the symbol
        SPY with `output_column_suffix` equal to MA30.

        `ta_func` is a wrapper for a technical analysis function.  The
        actual technical analysis function could be from ta-lib,
        pandas, pinkfish indicator, or a custom user function.
        `ta_param` is used to pass 1 parameter to `ta_func`.  Other
        parameters could be passed to the technical indicator within
        `ta_func`.  If you need to mass more than 1 paramters to
        `ta_func`, you could make `ta_param` a dict.

        Parameters
        ----------
        ts : pd.DataFrame
            The timeseries of the portfolio.
        ta_func : function
            A wrapper for a technical analysis function.
        ta_param : object
            The parameter for `ta_func` (typically an int).
        output_column_suffix : str
            Output column suffix to use for technical indicator.
        input_column_suffix : str, {'close', 'open', 'high', 'low'}
            Input column suffix to use for price (default is 'close').

        Returns
        -------
        ts : pd.DataFrame
            Timeseries with new column for technical indicator.

        Examples
        --------
        >>> # Add technical indicator: X day high
        >>> def period_high(ts, ta_param, input_column):
        >>>     return pd.Series(ts[input_column]).rolling(ta_param).max()

        >>> ts = portfolio.add_technical_indicator(
        >>>     ts, ta_func=_period_high, ta_param=period,
        >>>     output_column_suffix='period_high'+str(period),
        >>>     input_column_suffix='close')
        """
        indicator_column = {}
        for symbol in self.symbols:
            input_column = symbol + '_' + input_column_suffix
            output_column = symbol + '_' + output_column_suffix
            ts[output_column] = ta_func(ts, ta_param, input_column)
            indicator_column[output_column] = ta_func(ts, ta_param, input_column)
            
        # Join all the symbol columns to the original DataFrame using pd.concat
        ts = pd.concat([ts, pd.DataFrame(indicator_column)], axis=1)
        return ts

    def calendar(self, ts):
        """
        Add calendar columns to a timeseries.

        Parameters
        ----------
        ts : pd.DataFrame
            The timeseries of a symbol.

        Returns
        -------
        pd.DataFrame
            The timeseries with calendar columns added.
        """
        return calendar(ts)

    def finalize_timeseries(self, ts, start, dropna=True):
        """
        Finalize timeseries.

        Drop all rows that have nan column values.  Set timeseries to begin
        at start.

        Parameters
        ----------
        ts : pd.DataFrame
            The timeseries of a symbol.
        start : datetime.datetime
            The start date for backtest.
        dropna : bool, optional
            Drop rows that have a NaN value in one of it's columns
            (default is True).

        Returns
        -------
        datetime.datetime
            The start date.
        pd.DataFrame
            The timeseries of a symbol.
        """
        return finalize_timeseries(ts, start, dropna=dropna)

    ####################################################################
    # GET PRICES (get_price, get_prices)

    def get_column_value(self, row, symbol, field='close'):
        """
        Return column value given row, symbol, and field.

        Parameters
        ----------
        row : pd.Series
            The row of data from the timeseries of the portfolio.
        symbol : str
            The symbol for a security.
        field : str, optional {'close', 'open', 'high', 'low'}
            The price field to use (default is 'close').

        Returns
        -------
        price : float
            The current column value.
        """
        symbol += '_' + field
        try:
            price = getattr(row, symbol)
        except AttributeError:
            # This method is slower, but handles column names that
            # don't conform to variable name rules, and thus aren't
            # attributes.
            date = row.Index.to_pydatetime()
            price = self._ts.loc[date, symbol]
        return price
        
    get_price = get_column_value
    """
    method : get_price is a function reference to get_column_value, i.e. an alias.
    """

    def get_column_values(self, row, fields=['open', 'high', 'low', 'close']):
        """
        Return dict of column values for all symbols given row and fields.

        Parameters
        ----------
        row : pd.Series
            A row of data from the timeseries of the portfolio.
        fields : list, optional
            The list of fields to use for each symbol (default is
            ['open', 'high', 'low', 'close']).

        Returns
        -------
        d : dict of floats
            The column value indexed by symbol and field.
        """
        d = {}
        for symbol in self.symbols:
            d[symbol] = {}
            for field in fields:
                value = self.get_price(row, symbol, field)
                d[symbol][field] = value
        return d
        
    get_prices = get_column_values
    """
    method : get_prices is a function reference to get_column_values, i.e. an alias.
    """

    ####################################################################
    # ADJUST POSITION (adjust_shares, adjust_value, adjust_percent, print_holdings)

    def _share_value(self, row, field):
        """
        Return total share value in portfolio.
        """
        value = 0
        for symbol, tlog in trade.TradeLog.instance.items():
            price = self.get_price(row, symbol, field)
            value += tlog.share_value(price)
        return value

    def _total_value(self, row, field):
        """
        Return total_value = share_value + cash (if cash > 0).
        """
        total_value = self._share_value(row, field)
        if trade.TradeLog.cash > 0:
            total_value += trade.TradeLog.cash
        return total_value

    def _equity(self, row, field):
        """
        Return equity = total_value - loan (loan is negative cash)
        """
        equity = self._total_value(row, field)
        if trade.TradeLog.cash < 0:
            equity += trade.TradeLog.cash
        return equity

    def _leverage(self, row, field):
        """
        Return the leverage factor of the position.
        """
        return self._total_value(row, field) / self._equity(row, field)

    def _total_funds(self, row, field):
        """
        Return total account funds for trading.
        """
        return self._equity(row, field) * trade.TradeLog.margin

    def shares(self, symbol):
        """
        Return number of shares for given symbol in portfolio.

        Parameters
        ----------
        symbol : str
            The symbol for a security.

        Returns
        -------
        tlog.shares : int
            The number of shares for a given symbol.
        """
        tlog = trade.TradeLog.instance[symbol]
        return tlog.shares

    @property
    def positions(self):
        """
        Return the active symbols in portfolio as a list.

        This returns only those symbols that currently have shares
        allocated to them, either long or short.

        Parameters
        ----------
        None

        Returns
        -------
        list of str
            The active symbols in portfolio.
        """
        return [symbol for symbol in self.symbols if self.shares(symbol) > 0]

    def share_percent(self, row, symbol, field):
        """
        Return share value of symbol as a percentage of `total_funds`.

        Parameters
        ----------
        row : pd.Series
            A row of data from the timeseries of the portfolio.
        symbol : str
            The symbol for a security.
        field : str, {'close', 'open', 'high', 'low'}
            The price field to use.

        Returns
        -------
        float
            The share value as a percent.
        """
        price = self.get_price(row, symbol, field)
        tlog = trade.TradeLog.instance[symbol]
        value = tlog.share_value(price)
        return value / self._total_funds(row, field)

    def _calc_buying_power(self, row, field):
        """
        Return the buying power.
        """
        buying_power = (trade.TradeLog.cash * trade.TradeLog.margin
                      + self._share_value(row, field) * (trade.TradeLog.margin -1))
        return buying_power

    def _adjust_shares(self, row, price, shares, symbol, field, direction):
        """
        Adjust shares.
        """
        date = row.Index.to_pydatetime()
        tlog = trade.TradeLog.instance[symbol]
        trade.TradeLog.buying_power = self._calc_buying_power(row, field)
        shares = tlog.adjust_shares(date, price, shares, direction)
        trade.TradeLog.buying_power = None
        return shares

    def _adjust_value(self, row, value, symbol, field, direction):
        """
        Adjust value.
        """
        total_funds = self._total_funds(row, field)
        price = self.get_price(row, symbol, field)
        shares = int(min(total_funds, value) / price)
        shares = self._adjust_shares(row, price, shares, symbol, field, direction)
        return shares

    def adjust_percent(self, row, weight, symbol, field='close',
                       direction=trade.Direction.LONG):
        """
        Adjust symbol to a specified weight (percent) of portfolio.

        Parameters
        ----------
        row : pd.Series
            A row of data from the timeseries of the portfolio.
        weight : float
            The requested weight for the symbol, where 0 <= weight <=1.
        symbol : str
            The symbol for a security.
        field : str, {'close', 'open', 'high', 'low'}
            The price field to use.
        direction : pf.Direction, optional
            The direction of the trade (default is `pf.Direction.LONG`).

        Returns
        -------
        int
            The number of shares bought or sold.
        """
        if not (0 <= weight <= 1):
            raise ValueError(f'weight should be between 0 and 1 (inclusive), but {symbol}={weight}.')

        total_funds = self._total_funds(row, field)
        value = total_funds * weight
        shares = self._adjust_value(row, value, symbol, field, direction)
        return shares

    def adjust_percents(self, row, weights, field='close', directions=None):
        """
        Adjust symbols to a specified weight (percent) of portfolio.

        This function assumes all positions are LONG.  Prices and
        weights are given for all symbols in the portfolio.  The
        ordering of the prices and weights dicts are unimportant.
        They are dicts which are indexed by the symbol.

        Parameters
        ----------
        row : pd.Series
            A row of data from the timeseries of the portfolio.
        weights : dict of floats
            Dict of key value pair of symbol:weight, where 0 <= weight <=1.
        field : str, {'close', 'open', 'high', 'low'}
            The price field to use.
        directions : dict of pf.Direction, optional
            Dict of key value pair of symbol:direction.  The direction
            of the trades (default is None, which implies that all
            positions are long).

        Returns
        -------
        w : dict of floats
            Dict of key value pair of symbol:weight, where 0 <= weight <=1.
        """
        for symbol, weight in weights.items():
            if not (0 <= weight <= 1):
                raise ValueError(f'weights should be between 0 and 1 (inclusive), but {symbol}={weight}')

        w = {}

        # Get current weights
        for symbol in self.symbols:
            w[symbol] = self.share_percent(row, symbol, field)

        # If direction is None, this set all to pf.Direction.LONG
        if directions is None:
            directions = {symbol:trade.Direction.LONG for symbol in self.symbols}

        # We want to sell current positions first to obtain cash. We need to sort
        # the change of the current weight of the position to the new weight of the
        # position and order negative / sell orders first.
        for k, v in w.items():
            w[k] = weights[k] - v
        w = utility.sort_dict(w)

        # Update weights with new values.
        w.update(weights)

        # Call adjust_percent() for each symbol.
        for symbol, weight in w.items():
            direction = directions[symbol]
            self.adjust_percent(row, weight, symbol, field, direction)
        return w

    def print_holdings(self, row, show_percent=False):
        """
        Print snapshot of portfolio holding and values.

        Includes all symbols regardless of whether a symbol has shares
        currently allocated to it.

        Parameters
        ----------
        row : pd.Series
            A row of data from the timeseries of the portfolio.
        show_percent : bool, optional
            Show each holding as a percent instead of shares.
            (default is False).

        Returns
        -------
        None
        """
        date = row.Index.to_pydatetime()
        field = 'close'
        if show_percent:
            # 2007-11-20 SPY:24.1 TLT:24.9 GLD:24.6 QQQ:24.7 cash:  1.6 total: 100.0
            print(date.strftime('%Y-%m-%d'), end=' ')
            total = 0
            for symbol, tlog in trade.TradeLog.instance.items():
                pct = self.share_percent(row, symbol, field)
                total += pct
                print(f'{symbol}:{pct * 100:4,.1f}', end=' ')
            pct = trade.TradeLog.cash / self._equity(row, field)
            total += abs(pct)
            print(f'cash: {pct * 100:4,.1f}', end=' ')
            print(f'total: {total * 100:4,.1f}')
        else:
            # 2010-02-01 SPY: 54 TLT: 59 GLD:  9 cash:    84.20 total:  9,872.30
            print(date.strftime('%Y-%m-%d'), end=' ')
            for symbol, tlog in trade.TradeLog.instance.items():
                print(f'{symbol}:{tlog.shares:3}', end=' ')
            print(f'cash: {trade.TradeLog.cash:8,.2f}', end=' ')
            print(f'total: {self._equity(row, field):9,.2f}')

    ####################################################################
    # LOGS (init_trade_logs, record_daily_balance, get_logs)

    def init_trade_logs(self, ts):
        """
        Add a trade log for each symbol.

        Parameters
        ----------
        ts : pd.DataFrame
            The timeseries of the portfolio.

        Returns
        -------
        None
        """
        trade.TradeLog.seq_num = 0
        trade.TradeLog.instance.clear()

        self._ts = ts
        for symbol in self.symbols:
            trade.TradeLog(symbol, False)

    def record_daily_balance(self, row):
        """
        Append to daily balance list.

        The portfolio version of this function uses closing values
        for the daily high, low, and close.

        Parameters
        ----------
        row : pd.Series
            A row of data from the timeseries of the portfolio.

        Returns
        -------
        None
        """

        # calculate daily balance values: date, high, low, close,
        # shares, cash
        date = row.Index.to_pydatetime()
        field = 'close'
        equity = self._equity(row, field)
        leverage = self._leverage(row, field)
        shares = 0
        for tlog in trade.TradeLog.instance.values():
            shares += tlog.shares
        t = (date, equity, equity, equity, shares,
             trade.TradeLog.cash, leverage)
        self._l.append(t)

    def get_logs(self):
        """
        Return raw tradelog, tradelog, and daily balance log.

        Parameters
        ----------
        None

        Returns
        -------
        rlog : pd.DataFrame
            The raw trade log.
        tlog : pd.DataFrame
            The trade log.
        dbal : pd.DataFrame
            The daily balance log.
        """
        tlogs = []; rlogs = []
        for tlog in trade.TradeLog.instance.values():
            rlogs.append(tlog.get_log_raw())
            tlogs.append(tlog.get_log(merge_trades=False))
        
        rlogs_non_empty = [r for r in rlogs if not r.empty]
        rlog = pd.concat(rlogs_non_empty).sort_values(['seq_num'])
        
        tlogs_non_empty = [t for t in tlogs if not t.empty]
        tlog = pd.concat(tlogs_non_empty).sort_values(['entry_date', 'exit_date'])

        tlog['cumul_total'] = tlog['pl_cash'].cumsum()

        dbal = trade.DailyBal()
        dbal._l = self._l
        dbal = dbal.get_log(tlog)
        return rlog, tlog, dbal

    ####################################################################
    # PERFORMANCE ANALYSIS (performance_per_symbol, correlation_map)

    def performance_per_symbol(self, weights):
        """
        Returns performance per symbol data, also plots performance.

        Parameters
        ----------
        weights : dict of floats
            A dictionary of weights with symbol as key, where 0 <= weight <=1.

        Returns
        -------
        df : pd.DataFrame
            The dataframe contains performance for each symbol in the
            portfolio.
        """

        def _weight(row, weights):
            return weights[row.name]

        def _currency(row):
            return pfstatistics.currency(row['cumul_total'])

        def _plot(df):
            df = df[:-1]
            # Make new figure and set the size.
            fig = plt.figure(figsize=(12, 8))
            axes = fig.add_subplot(111, ylabel='Percentages')
            axes.set_title('Performance by Symbol')
            df.plot(kind='bar', ax=axes)
            axes.set_xticklabels(df.index, rotation=60)
            plt.legend(loc='best')

        for weight in weights.values():
            if not (0 <= weight <= 1):
                raise ValueError('weights should be between 0 and 1 (inclusive).')

        # Convert dict to series.
        s = pd.Series(dtype='object')
        for symbol, tlog in trade.TradeLog.instance.items():
            s[symbol] = tlog.cumul_total
        # Convert series to dataframe.
        df = pd.DataFrame(s.values, index=s.index, columns=['cumul_total'])
        # Add weight column.
        df['weight'] = df.apply(_weight, weights=weights, axis=1)
        # Add percent column.
        df['pct_cumul_total'] = df['cumul_total'] / df['cumul_total'].sum()
        # Add relative preformance.
        df['relative_performance'] = df['pct_cumul_total'] / df['weight']
        # Add TOTAL row.
        data = {'cumul_total':df['cumul_total'].sum(),
                'pct_cumul_total': 1.00, 'weight': 1.00,
                'relative_performance': 1.00}
        index = ['TOTAL']
        new_row = pd.DataFrame(data=data, index=index)
        df = pd.concat([df, new_row])
        # Format as currency.
        df['cumul_total'] = df.apply(_currency, axis=1)
        # Plot bar graph of performance.
        _plot(df)
        return df

    def correlation_map(self, ts, method='log', days=None):
        """
        Show correlation map between symbols.

        Parameters
        ----------
        ts : pd.DataFrame
            The timeseries of the portfolio.
        method : str, optional {'price', 'log', 'returns'}
            Analysis done based on specified method (default is 'log').
        days : int
            How many days to use for correlation (default is None,
            which implies all days).

        Returns
        -------
        df : pd.DataFrame
            The dataframe contains the correlation data for each symbol
            in the portfolio.
        """

        # Filter coloumn names for ''_close''; remove '_close' suffix.
        df = ts.filter(regex='_close')
        df.columns = df.columns.str.strip('_close')

        # Default is all days.
        if days is None:
            days = 0
        df = df[-days:]

        if method == 'price':
            pass
        elif method == 'log':
            df = np.log(df.pct_change()+1)
        elif method == 'returns':
            df = df.pct_change()

        df = df.corr(method='pearson')
        # Reset symbol as index (rather than 0-X).
        df.head().reset_index()
        # Take the bottom triangle since it repeats itself.
        mask = np.zeros_like(df)
        mask[np.triu_indices_from(mask)] = True
        # Generate plot.
        seaborn.heatmap(df, cmap='RdYlGn', vmax=1.0, vmin=-1.0,
                        mask=mask, linewidths=2.5)
        plt.yticks(rotation=0)
        plt.xticks(rotation=90)
        return df



"""
Adds calendar columns to a timeseries.

 - `dotw` : int, {0-6}
    Day of the week with Monday=0, Sunday=6.

 - `dotm` : int, {1,2,...}
    Day of the month as 1,2,...

 - `doty` : int, {1,2,...}
    Day of the year as 1,2,...

 - `month` : int, {1-12}
    Month as January=1,...,December=12

 - `first_dotw` : bool
    First trading day of the week.

 - `last_dotw` : bool
    Last trading day of the week.

 - `first_dotm` : bool
    First trading day of the month.

 - `last_dotm` : bool
    Last trading day of the month.

 - `first_doty` : bool
    First trading day of the year.

 - `last_doty` : bool
    Last trading day of the year.
"""


def _first_day(row):
    """
    Get the first trading day of week, month, year.
    """
    first_dotw = row['dotw'] < row['__prev_dotw__']
    first_dotm = row['dotm'] < row['__prev_dotm__']
    first_doty = row['doty'] < row['__prev_doty__']

    return first_dotw, first_dotm, first_doty


def calendar(ts):
    """
    Add calendar columns to a timeseries.

    Parameters
    ----------
    ts : pd.DataFrame
        The timeseries of a symbol.

    Returns
    -------
    pd.DataFrame
        The timeseries with calendar columns added.
    """

    # Day of the week with Monday=0, Sunday=6.
    ts['dotw'] = ts.index.dayofweek

    # Day of the month.
    ts['dotm'] = ts.index.day

    # Day of the year.
    ts['doty'] = ts.index.dayofyear

    # Month as January=1, December=12.
    ts['month'] = ts.index.month

    # Temporarily add __prev_dotw__, __prev_dotm__, __prev_doty__ for
    # convenience; drop them later.
    ts['__prev_dotw__'] = ts['dotw'].shift()
    ts['__prev_dotw__'].fillna(0, inplace=True)

    ts['__prev_dotm__'] = ts['dotm'].shift()
    ts['__prev_dotm__'].fillna(0, inplace=True)

    ts['__prev_doty__'] = ts['doty'].shift()
    ts['__prev_doty__'].fillna(0, inplace=True)

    # First and last day of the week, month, and year.
    ts['first_dotw'], ts['first_dotm'], ts['first_doty'] = \
        zip(*ts.apply(_first_day, axis=1))

    ts['last_dotw'] = ts['first_dotw'].shift(-1)
    ts['last_dotw'].fillna(False, inplace=True)

    ts['last_dotm'] = ts['first_dotm'].shift(-1)
    ts['last_dotm'].fillna(False, inplace=True)

    ts['last_doty'] = ts['first_doty'].shift(-1)
    ts['last_doty'].fillna(False, inplace=True)

    # Drop temporary columns.
    ts.drop(columns=['__prev_dotw__', '__prev_dotm__', '__prev_doty__'],
            inplace=True)

    return ts



"""
Fetch time series data.
"""

import datetime
import os
import sys
import warnings

import pandas as pd
from pandas_datareader._utils import RemoteDataError
import pandas_datareader.data as pdr
import yfinance as yf

from pinkfish.pfstatistics import (
    select_trading_days
)
from pinkfish.stock_market_calendar import (
    stock_market_calendar
)
import pinkfish.utility as utility


# Override pandas_datareader with yfinance
yf.pdr_override()


########################################################################
# TIMESERIES (fetch, select, finalize)

def _get_cache_dir(dir_name):
    """
    Get the data dir path.

    Parameters
    ----------
    dir_name : str
        The leaf data dir name.

    Returns
    -------
    str
        Path to the data dir.
    """
    base_dir = ''
    try:
        conf = utility.read_config()
        base_dir = conf['base_dir']
    except Exception as e:
        pass
    finally:
        dir_name = os.path.join(base_dir, dir_name)

    if not os.path.exists(dir_name):
        os.makedirs(dir_name)
    return dir_name


def _adj_column_names(ts):
    """
    Make all column names lower case.

    ta-lib expects columns to be lower case. To be consistent,
    make date index lowercase also.  Replace spaces with underscores.

    Parameters
    ----------
    ts : pd.DataFrame
        The timeseries of a symbol.

    Returns
    -------
    pd.DataFrame
        The timeseries with adjusted column names.
    """
    ts.columns = [col.lower().replace(' ','_') for col in ts.columns]
    ts.index.names = ['date']
    return ts


def fetch_timeseries(symbol, dir_name='data', use_cache=True, from_year=None):
    """
    Read time series data.

    Use cached version if it exists and use_cache is True, otherwise
    retrive, cache, then read.

    Parameters
    ----------
    symbol : str
        The symbol for a security.
    dir_name : str, optional
        The leaf data dir name (default is 'data').
    use_cache: bool, optional
        True to use data cache.  False to retrieve from the internet 
        (default is True).
    from_year: int, optional
        The start year for timeseries retrieval (default is None,
        which implies that all the available data is retrieved).

    Returns
    -------
    pd.DataFrame
        The timeseries of a symbol.
    """
    if from_year is None:
        from_year = 1900 if not sys.platform.startswith('win') else 1971

    # Yahoo finance uses '-' where '.' is used in symbol names.
    symbol = symbol.replace('.', '-')
    symbol = symbol.upper()

    # pinkfish allows the use of a suffix starting with a '_',
    # like SPY_SHRT, so extract the symbol.
    symbol = symbol.split('_')[0]

    timeseries_cache = os.path.join(_get_cache_dir(dir_name), symbol + '.csv')

    if os.path.isfile(timeseries_cache) and use_cache:
        pass
    else:
        try:
            ts = pdr.get_data_yahoo(symbol, start=datetime.datetime(from_year, 1, 1), progress=False)
        except RemoteDataError as e:
            print(f'\n{e}')
        except Exception as e:
            print(f'\n{e}')
        else:
            ts.to_csv(timeseries_cache, encoding='utf-8')

    ts = pd.read_csv(timeseries_cache, index_col='Date', parse_dates=True)
    ts = _adj_column_names(ts)

    # Remove rows that have duplicated index.
    ts = ts[~ts.index.duplicated(keep='first')]
    return ts


def _adj_prices(ts):
    """
    Back adjust prices relative to adj_close for dividends and splits.

    Parameters
    ----------
    ts : pd.DataFrame
        The timeseries of a symbol.

    Returns
    -------
    pd.DataFrame
        The timeseries with adjusted prices.
    """
    ts['open'] = ts['open'] * ts['adj_close'] / ts['close']
    ts['high'] = ts['high'] * ts['adj_close'] / ts['close']
    ts['low'] = ts['low'] * ts['adj_close'] / ts['close']
    ts['close'] = ts['close'] * ts['adj_close'] / ts['close']
    return ts


def select_tradeperiod(ts, start, end, use_adj=False,
                       use_continuous_calendar=False,
                       force_stock_market_calendar=False,
                       check_fields=['close']):
    """
    Select the trade period.

    First, remove rows that have zero values in price columns. Then,
    select a time slice of the data to trade from ts.  Back date a year
    to allow time for long term indicators, e.g. 200sma is become valid.

    Parameters
    ----------
    ts : pd.DataFrame
        The timeseries of a symbol.
    start : datetime.datetime
        The desired start date for the strategy.
    end : datetime.datetime
        The desired end date for the strategy.
    use_adj : bool, optional
        True to adjust prices for dividends and splits
        (default is False).
    use_continuous_calendar: bool, optional
        True if your timeseries has data for all seven days a week,
        and you want to backtest trading every day, including weekends.
        If this value is True, then `force_stock_market_calendar`
        is set to False (default is False).
    force_stock_market_calendar : bool, optional
        True forces use of stock market calendar on timeseries.
        Normally, you don't need to do this.  This setting is intended
        to transform a continuous timeseries into a weekday timeseries.
        If this value is True, then `use_continuous_calendar` is set
        to False (default is False).
    check_fields : list of str, optional {'high', 'low', 'open',
        'close', 'adj_close'}
        Fields to check for for NaN values.  If a NaN value is found
        for one of these fields, that row is dropped
        (default is ['close']).

    Returns
    -------
    pd.DataFrame
        The timeseries for specified start:end, optionally with prices
        adjusted.

    Notes
    -----
    You should only set one of `use_continuous_calendar`=True or
    `force_stock_market_calendar`=True for a continuous timeseries.
    You should set neither of these to True if your timeseries is based
    on the stock market.
    """
    columns = ['high', 'low', 'open', 'close']
    if use_adj:
        columns.append('adj_close')

    # Replace 0 value columns with NaN.
    ts[columns] = ts[ts[columns] > 0][columns]

    if use_continuous_calendar:
        force_stock_market_calendar = False
    if force_stock_market_calendar:
        use_continuous_calendar = False

    if use_continuous_calendar:
        select_trading_days(use_stock_market_calendar=False)

    if force_stock_market_calendar:
        index = pd.to_datetime(stock_market_calendar)
        ts = ts.reindex(index=index)

    ts.dropna(subset=check_fields, inplace=True)

    if use_adj:
        _adj_prices(ts)

    if start < ts.index[0]:
        start = ts.index[0]
    if end > ts.index[-1]:
        end = ts.index[-1]
    ts = ts[start - datetime.timedelta(365):end]

    return ts


def finalize_timeseries(ts, start, dropna=False, drop_columns=None):
    """
    Finalize timeseries.

    Drop all rows that have nan column values.  Set timeseries to begin
    at start.

    Parameters
    ----------
    ts : pd.DataFrame
        The timeseries of a symbol.
    start : datetime.datetime
        The start date for backtest.
    dropna : bool, optional
        Drop rows that have a NaN value in one of it's columns
        (default is False).
    drop_columns : list of str, optional
        List of columns to drop from `ts` (default is None, which
        implies that no columns should be dropped).

    Returns
    -------
    datetime.datetime
        The start date.
    pd.DataFrame
        The timeseries of a symbol.
    """
    if drop_columns:
        ts.drop(columns=drop_columns, inplace=True)
    if dropna:
        ts.dropna(inplace=True)
    elif ts.isnull().values.any():
        warnings.warn("NaN value(s) detected in timeseries")
    ts = ts[start:]
    start = ts.index[0]
    return ts, start


#####################################################################
# CACHE SYMBOLS (remove, update, get_symbol_metadata)

def _difference_in_years(start, end):
    """
    Calculate the number of years between two dates.

    Parameters
    ----------
    start : datetime.datetime
        The start date.
    end : datetime.datetime
        The end date.

    Returns
    -------
    float
        The difference in years between start and end dates.
    """
    diff = end - start
    diff_in_years = (diff.days + diff.seconds/86400)/365.2425
    return diff_in_years


def remove_cache_symbols(symbols=None, dir_name='data'):
    """
    Remove cached timeseries for list of symbols.

    Filter out any symbols prefixed with '__'.

    Parameters
    ----------
    symbols : str or list of str, optional
        The symbol(s) for which to remove cached timeseries (default
        is None, which imples remove timeseries for all symbols).
    dir_name : str, optional
        The leaf data dir name (default is 'data').

    Returns
    -------
    None
    """
    cache_dir = _get_cache_dir(dir_name)

    if symbols:
        # If symbols is not a list, cast it to a list.
        if not isinstance(symbols, list):
            symbols = [symbols]
        filenames = [symbol.upper() + '.csv' for symbol in symbols]
    else:
        filenames = [f for f in os.listdir(cache_dir) if f.endswith('.csv')]

    # Filter out any filename prefixed with '__'.
    filenames = [f for f in filenames if not f.startswith('__')]

    print('removing symbols:')
    for i, f in enumerate(filenames):
        symbol = os.path.splitext(f)[0]
        print(symbol + ' ', end='')
        if i % 10 == 0 and i != 0: print()

        filepath = os.path.join(cache_dir, f)
        if os.path.exists(filepath):
            os.remove(filepath)
        else:
            print(f'\n({f} not found)')
    print()


def update_cache_symbols(symbols=None, dir_name='data', from_year=None):
    """
    Update cached timeseries for list of symbols.

    Filter out any filename prefixed with '__'.

    Parameters
    ----------
    symbols : str or list, optional
        The symbol(s) for which to update cached timeseries (default
        is None, which imples update timeseries for all symbols).
    dir_name : str, optional
        The leaf data dir name (default is 'data).
    from_year: int, optional
        The start year for timeseries retrieval (default is None,
        which implies that all the available data is retrieved).

    Returns
    -------
    None
    """
    cache_dir = _get_cache_dir(dir_name)

    if symbols:
        # If symbols is not a list, cast it to a list.
        if not isinstance(symbols, list):
            symbols = [symbols]
    else:
        filenames = ([f for f in os.listdir(cache_dir)
                      if f.endswith('.csv') and not f.startswith('__')])
        symbols = [os.path.splitext(filename)[0] for filename in filenames]

    # Make symbol names uppercase.
    symbols = [symbol.upper() for symbol in symbols]

    print('updating symbols:')
    for i, symbol in enumerate(symbols):
        print(symbol + ' ', end='')
        if i % 10 == 0 and i != 0:
            print()

        try:
            fetch_timeseries(symbol, dir_name=dir_name, use_cache=False,
                             from_year=from_year)
        except RemoteDataError as e:
            print(f'\n({e})')
        except Exception as e:
            print(f'\n({e})')
    print()


def get_symbol_metadata(symbols=None, dir_name='data', from_year=None):
    """
    Get symbol metadata for list of symbols.

    Filter out any filename prefixed with '__'.

    Parameters
    ----------
    symbols : str or list, optional
        The symbol(s) for which to get symbol metadata (default
        is None, which imples get symbol metadata for all symbols).
    dir_name : str, optional
        The leaf data dir name (default is 'data).
    from_year: int, optional
        The start year for timeseries retrieval (default is None,
        which implies that all the available data is retrieved).

    Returns
    -------
    pd.DataFrame
        Each row contains metadata for a symbol.
    """
    cache_dir = _get_cache_dir(dir_name)

    if symbols:
        # If symbols is not a list, cast it to a list.
        if not isinstance(symbols, list):
            symbols = [symbols]
    else:
        filenames = ([f for f in os.listdir(cache_dir)
                     if f.endswith('.csv') and not f.startswith('__')])
        symbols = [os.path.splitext(filename)[0] for filename in filenames]

    # Make symbol names uppercase.
    symbols = [symbol.upper() for symbol in symbols]

    l = []
    for i, symbol in enumerate(symbols):
        try:
            ts = fetch_timeseries(symbol, dir_name=dir_name, use_cache=True,
                                  from_year=from_year) 
            start = ts.index[0].to_pydatetime()
            end = ts.index[-1].to_pydatetime()
            num_years = _difference_in_years(start, end)
            start = start.strftime('%Y-%m-%d')
            end = end.strftime('%Y-%m-%d')
            t = (symbol, start, end, num_years)
            l.append(t)
        except RemoteDataError as e:
            print(f'\n({e})')
        except Exception as e:
            print('\n({})'.format(e))
    columns = ['symbol', 'start_date', 'end_date', 'num_years']
    df = pd.DataFrame(l, columns=columns)
    return df



'''
Keep track of styles for cells/headers in PrettyTable.

The MIT License (MIT)

Copyright (c) 2014 Melissa Gymrek <mgymrek@mit.edu>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
'''


class TableStyle(object):

    """
    Keep track of styles for cells/headers in PrettyTable
    """

    def __init__(self, theme=None):
        self.row_head_style = CellStyle()
        self.col_head_style = CellStyle()
        self.cell_style = CellStyle()
        self.corner_style = CellStyle()

        # add themes as needed
        if theme == "basic":
            self.cell_style.set("border", "1px solid black")
            self.col_head_style.set("font-weight", "bold")
            self.row_head_style.set("font-weight", "bold")

        if theme == "theme1":
            self.cell_style.set("border", "1px solid black")
            self.cell_style.set("color", "black")
            self.col_head_style.set("color", "black")
            self.row_head_style.set("color", "black")
            self.col_head_style.set("font-weight", "bold")
            self.row_head_style.set("font_weight", "bold")
            self.col_head_style.set("background-color", "lightgray")
            self.row_head_style.set("background-color", "lightgray")


class CellStyle(object):

    """
    Styles for cells PrettyTable
    """

    def __init__(self):
        self.style_elements = {}  # dictionary of CSS property -> value
        self.format_function = None

    def set(self, key, value):
        self.style_elements[key] = value

    def css(self):
        style = ""
        for key in self.style_elements:
            style += "%s: %s;" % (key, self.style_elements[key])
        return style

    def column_format(self, x):
        if self.format_function is None:
            return str(x)
        else:
            try:
                return self.format_function(x)
            except:
                return str(x)

    def copy(self):
        c = CellStyle()
        c.style_elements = self.style_elements.copy()
        c.format_function = self.format_function
        return c


class PrettyTable(object):

    """
    Formatted tables for display in IPython notebooks
    """

    def __init__(self, df, tstyle=None, header_row=False, header_col=True,
                 center=False, rpt_header=0):
        """
        df: pandas.DataFrame
        style: TableStyle
        header_row: include row headers
        header_col: include column headers
        """
        self.df = df
        self.num_rows = df.shape[0]
        self.num_cols = df.shape[1]
        self.header_row = header_row
        self.header_col = header_col
        self.style = tstyle
        self.center = center
        self.rpt_header = rpt_header

        # overall table style
        if tstyle is None:
            self.cell_style = CellStyle()
            self.corner_style = CellStyle()
            self.header_row_styles = [CellStyle()
                                      for i in range(self.num_rows)]
            self.header_col_styles = [CellStyle()
                                      for i in range(self.num_cols)]
            self.cell_styles = [[CellStyle() for i in range(self.num_cols)]
                                for j in range(self.num_rows)]
        else:
            self.cell_style = tstyle.cell_style
            self.corner_style = tstyle.corner_style
            self.header_row_styles = [
                tstyle.row_head_style.copy() for i in range(self.num_rows)]
            self.header_col_styles = [
                tstyle.col_head_style.copy() for i in range(self.num_cols)]
            self.cell_styles = [[self.cell_style.copy() for i in range(self.num_cols)]
                                for j in range(self.num_rows)]

    # functions to set styles
    def set_cell_style(self, style=None, tuples=None, rows=None, cols=None,
                       format_function=None, **kwargs):
        """
        Apply cell style to rows and columns specified
        """
        if style is None:
            style = CellStyle()
        for key, value in kwargs.items():
            k = key.replace("_", "-")
            style.set(k, value)
        if format_function is not None:
            style.format_function = format_function
        if tuples:
            for tup in tuples:
                i = tup[0]
                j = tup[1]
                self.cell_styles[i][j] = style.copy()
        if rows is None and cols is None:
            return
        if rows is None:
            rows = range(self.num_rows)
        if cols is None:
            cols = range(self.num_cols)
        for i in rows:
            for j in cols:
                self.cell_styles[i][j] = style.copy()

    def set_row_header_style(self, style=None, indices=None, format_function=None, **kwargs):
        """
        Apply style to header at specific index
        If index is None, apply to all headings
        """
        if style is None:
            style = CellStyle()
        for key, value in kwargs.items():
            k = key.replace("_", "-")
            style.set(k, value)
        if format_function is not None:
            style.format_function = format_function
        if indices is None:
            indices = range(self.num_rows)
        for i in indices:
            self.header_row_styles[i] = style.copy()

    def set_col_header_style(self, style=None, indices=None, format_function=None, **kwargs):
        """
        Apply style to header at specific index
        If index is None, apply to all headings
        """
        if indices is None:
            indices = range(self.num_cols)
        if style is None:
            style = CellStyle()
        if format_function is not None:
            style.format_function = format_function
        for key, value in kwargs.items():
            k = key.replace("_", "-")
            style.set(k, value)
        for i in indices:
            self.header_col_styles[i] = style.copy()

    def set_corner_style(self, style=None, format_function=None, **kwargs):
        """
        Apply style to the corner cell
        """
        if style is None:
            style = CellStyle()
        for key, value in kwargs.items():
            k = key.replace("_", "-")
            style.set(k, value)
        if format_function is not None:
            style.format_function = format_function
        self.corner_style = style

    # functions to update styles
    def update_cell_style(self, rows=None, cols=None, format_function=None, **kwargs):
        """
        Update existing cell style
        """
        if rows is None:
            rows = range(self.num_rows)
        if cols is None:
            cols = range(self.num_cols)
        for i in rows:
            for j in cols:
                style = self.cell_styles[i][j]
                self.set_cell_style(
                    style=style, rows=[i], cols=[j], format_function=format_function, **kwargs)

    def update_row_header_style(self, indices=None, format_function=None, **kwargs):
        """
        Update existing row header tyle
        """
        if indices is None:
            indices = range(self.num_rows)
        for i in indices:
            style = self.header_row_styles[i]
            self.set_row_header_style(
                style=style, indices=[i], format_function=format_function, **kwargs)

    def update_col_header_style(self, indices=None, format_function=None, **kwargs):
        """
        Update existing row header tyle
        """
        if indices is None:
            indices = range(self.num_cols)
        for i in indices:
            style = self.header_col_styles[i]
            self.set_col_header_style(
                style=style, indices=[i], format_function=format_function, **kwargs)

    def update_corner_style(self, format_function=None, **kwargs):
        """
        Update the corner style
        """
        style = self.corner_style
        self.set_corner_style(
            style=style, format_function=format_function, **kwargs)

    # Functions to reset style
    def reset_cell_style(self, rows=None, cols=None):
        """
        Reset existing cell style to defaults
        """
        if rows is None:
            rows = range(self.num_rows)
        if cols is None:
            cols = range(self.num_cols)
        for i in rows:
            for j in cols:
                self.set_cell_style(style=CellStyle(), rows=[i], cols=[j])

    def reset_row_header_style(self, indices=None):
        """
        Reset row header style to defaults
        """
        if indices is None:
            indices = range(self.num_rows)
        for i in indices:
            self.set_row_header_style(style=CellStyle(), indices=[i])

    def reset_col_header_style(self, indices=None):
        """
        Reset col header style to defaults
        """
        if indices is None:
            indices = range(self.num_cols)
        for i in indices:
            self.set_col_header_style(style=CellStyle(), indices=[i])

    def reset_corner_style(self):
        """
        Reset corner style to defaults
        """
        self.set_corner_style(style=CellStyle())

    def _repr_html_(self):
        """
        IPython display protocol calls this method to get the
        HTML representation of the object
        """
        html = "<table style=\"%s\">" % self.cell_style.css()
        if self.header_col:
            html += "<tr style=\"%s\">" % self.cell_style.css()
            if self.header_row:
                # need to add an extra empty cell
                html += "<td style=\"%s\"></td>" % self.corner_style.css()
            for j in range(self.num_cols):
                if self.header_col_styles is not None:
                    header_style = self.header_col_styles[j].css()
                    header_data = self.header_col_styles[
                        j].column_format(self.df.columns[j])
                else:
                    header_style = self.cell_style.css()
                    header_data = self.cell_style.column_format(
                        self.df.columns[j])
                html += "<td style=\"%s\">" % header_style
                html += header_data
                html += "</td>"
            html += "</tr>"
        for i in range(self.num_rows):
            html += "<tr style=\"%s\">" % self.cell_style.css()
            if self.header_row:
                if self.header_row_styles is not None:
                    header_style = self.header_row_styles[i].css()
                    header_data = self.header_row_styles[
                        i].column_format(self.df.index.values[i])
                else:
                    header_style = self.cell_style.css()
                    header_data = self.cell_style.column_format(
                        self.df.index.values[i])
                html += "<td style=\"%s\">" % header_style
                html += header_data
                html += "</td>"
            for j in range(self.num_cols):
                if self.cell_styles[i][j] is not None:
                    col_style = self.cell_styles[i][j].css()
                    col_data = self.cell_styles[i][
                        j].column_format(self.df.iloc[i, j])
                else:
                    col_style = self.cell_style.css()
                    col_data = self.cell_style.column_format(
                        self.df.iloc[i, j])
                html += "<td style=\"%s\">" % col_style
                html += col_data
                html += "</td>"
            html += "</tr>"
            if self.rpt_header > 0 and (i + 1) % self.rpt_header == 0 and i < self.num_rows - 1:
                if self.header_col:
                    html += "<tr style=\"%s\">" % self.cell_style.css()
                    if self.header_row:
                        # need to add an extra empty cell
                        html += "<td style=\"%s\"></td>" % self.corner_style.css(
                        )
                    for j in range(self.num_cols):
                        if self.header_col_styles is not None:
                            header_style = self.header_col_styles[j].css()
                            header_data = self.header_col_styles[
                                j].column_format(self.df.columns[j])
                        else:
                            header_style = self.cell_style.css()
                            header_data = self.cell_style.column_format(
                                self.df.columns[j])
                        html += "<td style=\"%s\">" % header_style
                        html += header_data
                        html += "</td>"
                    html += "</tr>"
        html += "</table>"
        if self.center:
            return "<center>%s</center>" % html
        else:
            return html

    def copy(self):
        p = PrettyTable(self.df, self.style, self.header_row, self.header_col)
        p.header_row_styles = [item.copy() for item in self.header_row_styles]
        p.header_col_styles = [item.copy() for item in self.header_col_styles]
        p.cell_styles = [[self.cell_styles[i][j].copy() for j in range(
            self.num_cols)] for i in range(self.num_rows)]
        p.corner_style = self.corner_style.copy()
        p.center = self.center
        return p



from setuptools import setup

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name='pinkfish',
    version='2.1.2',
    description='A backtester and spreadsheet library for security analysis.',
    author='Farrell Aultman',
    author_email='fja0568@gmail.com',
    url='https://github.com/fja05680/pinkfish',
    packages=['pinkfish'],
    include_package_data=True,
    license='MIT',
    install_requires=requirements,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
        'Intended Audience :: Science/Research',
        'Topic :: Office/Business :: Financial',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: System :: Distributed Computing',
    ],
    extras_require={
        'talib':  ['TA-Lib'],
    },
    data_files=[('', ['requirements.txt'])],
    python_requires=">=3.7",
)



"""
Trading agent.
"""

import pandas as pd


class Direction:
    """
    The direction of the trade.  Either LONG or SHORT.
    """
    LONG, SHORT = ['LONG', 'SHRT']

class Margin:
    """
    The type of margin.  CASH, STANDARD, or PATTERN_DAY_TRADER.
    """
    CASH, STANDARD, PATTERN_DAY_TRADER = [1, 2, 4]


########################################################################
# TRADE LOG - each symbol has it's own trade log

class TradeLog:
    """
    The trade log for each symbol.
    """

    cash = 0
    """
    int : Current cash, entire portfolio.
    """
    multiplier = 1
    """
    int : Applied to profit calculation.  Used only with futures.
    """
    margin = Margin.CASH
    """
    float : Margin percent.
    """
    buying_power = None
    """
    float : Buying power for Portfolio class.
    """
    seq_num = 0
    """
    int : Sequential number used to order trades in Portfolio class.
    """
    instance = {}
    """
    dict of pf.TradeLog : dict (key=symbol) of TradeLog instances used
    in Portfolio class.
    """

    def __init__(self, symbol, reset=True):
        """
        Initialize instance variables.

        Parameters
        ----------
        symbol : str
            The symbol for a security.
        reset : bool, optional
            Use when starting new portfolio construction to clear the
            dict of TradeLog instances (default is True).

        Attributes
        ----------
        symbol : str
            The symbol for a security.
        shares : int
            Number of shares of the symbol.
        direction : pf.Direction
            The direction of the trade, Long or Short.
        ave_entry_price : float
            The average purchase price per share.
        cumul_total : float
            The cumulative total profits (loss).
        _l : list of tuples
            The list of matching entry/exit trade pairs.  This list
            will become the official trade log.
        _raw : list of tuples
            The list of raw trades, either entry or exit.
        open_trades : list
            The list of open trades, i.e. not closed out.
        """
        self.symbol = symbol
        self.shares = 0
        self.direction = None
        self.ave_entry_price = 0
        self.cumul_total = 0
        self._l = []
        self._raw = []
        self._open_trades = []

        if reset:
            TradeLog.seq_num = 0
            TradeLog.instance.clear()
        TradeLog.instance[symbol] = self

    def share_value(self, price):
        """
        Return the total value of shares of the security.

        Parameters
        ----------
        price : float
            The current price of the security.

        Returns
        -------
        value : float
            The share value.
        """
        value = 0
        if self.direction == Direction.LONG:
            value += price*self.shares
        elif self.direction == Direction.SHORT:
            value += (2*self.ave_entry_price-price)*self.shares
        return value
        
    def total_value(self, price):
        """
        Return the total value which is the total share value plus cash.

        Parameters
        ----------
        price : float
            The current price of the security.

        Returns
        -------
        value : float
            The total value.
        """
        total_value = self.share_value(price)
        if TradeLog.cash > 0:
            total_value += TradeLog.cash
        return total_value

    def equity(self, price):
        """
        Return the equity which is the total value minus loan.
        Loan is negative cash.
        """
        equity = self.total_value(price)
        if TradeLog.cash < 0:
            equity += TradeLog.cash
        return equity

    def leverage(self, price):
        """
        Return the leverage factor of the position given current price.
        """
        return self.total_value(price) / self.equity(price)

    def total_funds(self, price):
        """
        Return the total account funds for trading given current price.
        """
        return self.equity(price) * TradeLog.margin

    def share_percent(self, price):
        """
        Return the share value as a percentage of total funds.
        """
        return self.share_value(price) / self.total_funds(price)

    @property
    def num_open_trades(self):
        """
        Return the number of open orders, i.e. not closed out.
        """
        return len(self._open_trades)

    ####################################################################
    # ENTER TRADE (buy, sell_short)

    def calc_buying_power(self, price):
        """
        Calculate buying power.
        """
        buying_power = (TradeLog.cash * TradeLog.margin
                        + self.share_value(price) * (TradeLog.margin -1))
        return buying_power

    def calc_shares(self, price, cash=None):
        """
        Calculate shares using buying power before enter_trade().

        Parameters
        ----------
        price : float
            The current price of the security.
        cash : float, optional
            The requested amount of cash used to buy shares (default
            is None, which implies use all available cash).

        Returns
        -------
        value : float
            The number of shares that can be purchased with requested
            cash amount.
        """

        # Margin should be equal to or greater than 1.
        if TradeLog.margin < 1: TradeLog.margin = 1

        # Calculate buying power.  TradeLog.buying_power may have
        # already been calculated in portfolio.
        if TradeLog.buying_power is not None:
            buying_power = TradeLog.buying_power
        else:
            buying_power = self.calc_buying_power(price)

        # Cash can't exceed buying power.
        if cash is None or cash > buying_power:
            cash = buying_power

        # Cash can't be negative.
        if cash < 0: cash = 0

        # Calculate shares.
        shares = int(cash / price)
        return shares

    def _enter_trade(self, entry_date, entry_price, shares=None, direction=Direction.LONG):
        """
        This is a lower level function that gets called from
        enter_trade() and sell_short().
        """

        max_shares = self.calc_shares(entry_price)
        shares = max_shares if shares is None else min(shares, max_shares)

        if shares == 0:
            return 0

        # Record in raw trade log.
        t = (entry_date, TradeLog.seq_num, entry_price, shares, 'entry', direction, self.symbol)
        self._raw.append(t)
        TradeLog.seq_num += 1

        # Add record to open_trades.
        d = {'entry_date':entry_date, 'entry_price':entry_price, 'qty':shares,
             'direction':direction, 'symbol':self.symbol}
        self._open_trades.append(d)

        # Update direction.
        if self.direction != direction:
            if self.direction is None or self.shares == 0:
                self.direction = direction
            else:
                raise ValueError('not allowed to change direction from {} to {}, '
                                 'this requires shares = 0'
                                 .format(self.direction, direction))

        # Update average entry price and shares.
        self.ave_entry_price = \
            (self.ave_entry_price*self.shares + entry_price*shares) / (self.shares + shares)
        self.shares += shares

        # Update cash.
        TradeLog.cash -= entry_price * shares

        return shares

    def enter_trade(self, entry_date, entry_price, shares=None):
        """
        Enter a trade on the long side.

        Parameters
        ----------
        entry_date : str
            The entry date.
        entry_price : float
            The entry price.
        shares : int, optional
            The number of shares to buy (default is None, which implies
            buy the maximum number of shares possible with available
            buying power).

        Returns
        -------
        int
            The number of shares bought.

        Notes
        -----
        The `buy' alias can be used to call this function for increasing
        or opening a long position.
        """
        return self._enter_trade(entry_date=entry_date,
                                 entry_price=entry_price,
                                 shares=shares,
                                 direction=Direction.LONG)

    buy = enter_trade
    """
    method : buy is a function reference to enter_trade, i.e. an alias.
    """

    def sell_short(self, entry_date, entry_price, shares=None):
        """
        Enter a trade on the short side.

        Parameters
        ----------
        entry_date : str
            The entry date.
        entry_price : float
            The entry price.
        shares : int
            The number of shares to sell short (default in None, which
            implies to sell short the maximum number of shares
            possible).

        Returns
        -------
        int
            The number of shares sold short.
        """
        return self._enter_trade(entry_date=entry_date,
                                 entry_price=entry_price,
                                 shares=shares,
                                 direction=Direction.SHORT)

    ####################################################################
    # EXIT TRADE (sell, buy2cover)

    def _qty_open_trade(self, index):
        """
        Qty of an open trade by index.
        """
        if index >= self.num_open_trades:
            return 0
        return self._open_trades[index]['qty']

    def _exit_trade(self, exit_date, exit_price, shares=None, direction=Direction.LONG):
        """
        Exit a trade.

        Record exit in trade log. return -shares exited.
        shares = None exits all shares
        shares > 0 exits that number of shares
        shares < 0 indicates the number of open_trades to exit
        """
        if shares is None or shares > self.shares:
            shares = self.shares
        elif shares < 0:
            open_trades = -shares
            shares = 0
            for i in range(open_trades):
                shares += self._qty_open_trade(i)

        if shares == 0:
            return 0

        shares_orig = shares

        # Record in raw trade log.
        t = (exit_date, TradeLog.seq_num, exit_price, shares, 'exit', direction, self.symbol)
        self._raw.append(t)
        TradeLog.seq_num += 1

        for i, open_trade in enumerate(self._open_trades[:]):
            entry_date = open_trade['entry_date']
            entry_price = open_trade['entry_price']
            qty = open_trade['qty']

            if direction == Direction.LONG:
                pl_points = exit_price - entry_price
            else:
                pl_points = -(exit_price - entry_price)

            # Calculate exit_shares and pl_cash.
            exit_shares = qty if shares >= qty else shares
            pl_cash = pl_points * exit_shares * TradeLog.multiplier
            self.cumul_total += pl_cash

            # Record in trade log.
            t = (entry_date, entry_price, exit_date, exit_price,
                 pl_points, pl_cash, exit_shares, self.cumul_total,
                 direction, self.symbol)
            self._l.append(t)

            # Update shares and cash.
            self.shares -= exit_shares
            TradeLog.cash += self.ave_entry_price*exit_shares + pl_cash

            # Update open_trades list.
            if shares == qty:
                del self._open_trades[0]
                break
            elif shares < qty:
                self._open_trades[0]['qty'] -= shares
                break
            else:
                del self._open_trades[0]
                shares -= exit_shares

        return -shares_orig

    def exit_trade(self, exit_date, exit_price, shares=None):
        """
        Exit a trade on the long side.

        Parameters
        ----------
        exit_date : str
            The exit date.
        exit_price : float
            The exit price.
        shares : int, optional
            The number of shares to sell (default is None, which implies
            sell all the shares).

        Returns
        -------
        int
            The number of shares sold.

        Notes
        -----
        The `sell' alias can be used to call this function for reducing
        or closing out a long position.
        """
        return self._exit_trade(exit_date=exit_date,
                                exit_price=exit_price,
                                shares=shares,
                                direction=Direction.LONG)

    sell = exit_trade
    """
    method : sell is a function reference to exit_trade, i.e. an alias.
    """

    def buy2cover(self, exit_date, exit_price, shares=None):
        """
        Exit a trade on the short side, i.e. buy to cover.

        Parameters
        ----------
        exit_date : str
            The exit date.
        exit_price : float
            The exit price.
        shares : int
            The number of shares to buy to cover (default in None,
            which implies close out the short shares).

        Returns
        -------
        int
            The number of shares bought.
        """
        return self._exit_trade(exit_date=exit_date,
                                exit_price=exit_price,
                                shares=shares,
                                direction=Direction.SHORT)

    ####################################################################
    # GET PRICES (get_price, get_prices)

    def get_price(self, row, field='close'):
        """
        Return price given row and field.

        Parameters
        ----------
        row : pd.Series
            The timeseries of the portfolio.
        field : str, optional {'close', 'open', 'high', 'low'}
            The price field (default is 'close').

        Returns
        -------
        float
            The current price.

        """
        return getattr(row, field)

    def get_prices(self, row, fields=['open', 'high', 'low', 'close']):
        """
        Return dict of prices for all symbols given row and fields.

        Parameters
        ----------
        row : pd.Series
            The timeseries of the portfolio.
        fields : list, optional
            The list of fields to use (default is
            ['open', 'high', 'low', 'close']).

        Returns
        -------
        d : dict of floats
            The price indexed by fields.
        """
        d = {}
        for field in fields:
            value = self.get_price(row, field)
            d[field] = value
        return d

    ####################################################################
    # ADJUST POSITION (adjust_shares, adjust_value, adjust_percent)

    def adjust_shares(self, date, price, shares, direction=Direction.LONG):
        """
        Adjust a position to a target number of shares.

        If the position doesn't already exist, this is equivalent
        to entering a new trade. If the position does exist, this is
        equivalent to entering or exiting a trade for the difference
        between the target number of shares and the current number
        of shares.

        Parameters
        ----------
        date : str
            The trade date.
        price : float
            The current price of the security.
        shares : int
            The requested number of target shares.
        direction : pf.Direction, optional
            The direction of the trade (default is Direction.LONG).

        Returns
        -------
        int
            The number of shares bought or sold.
        """
        diff_shares = shares - self.shares
        if direction == Direction.LONG:
            if diff_shares >= 0:
                shares = self.enter_trade(date, price, diff_shares)
            else:
                shares = self.exit_trade(date, price, -diff_shares)
        else:
            if diff_shares >= 0:
                shares = self.sell_short(date, price, diff_shares)
            else:
                shares = self.buy2cover(date, price, -diff_shares)
        return shares

    def adjust_value(self, date, price, value, direction=Direction.LONG):
        """
        Adjust a position to a target value.

        If the position doesn't already exist, this is equivalent
        to entering a new trade. If the position does exist, this is
        equivalent to entering or exiting a trade for the difference
        between the target value and the current value.

        Parameters
        ----------
        date : str
            The trade date.
        price : float
            The current price of the security.
        shares : int
            The requested target value.
        direction : pf.Direction, optional
            The direction of the trade (default is Direction.LONG).

        Returns
        -------
        int
            The number of shares bought or sold.
        """
        total_funds = self.total_funds(price)
        shares = int(min(total_funds, value) / price)
        shares = self.adjust_shares(date, price, shares, direction)
        return shares

    def adjust_percent(self, date, price, weight, direction=Direction.LONG):
        """
        Adjust position to a target percent of the current portfolio
        value.

        If the position doesn't already exist, this is equivalent
        to entering a new trade. If the position does exist, this is
        equivalent to entering or exiting a trade for the difference
        between the target percent and the current percent.

        Parameters
        ----------
        date : str
            The trade date.
        price : float
            The current price of the security.
        weight : int
            The requested target weight, , where 0 <= weight <= 1.
        direction : pf.Direction, optional
            The direction of the trade (default is Direction.LONG).

        Returns
        -------
        int
            The number of shares bought or sold.
        """
        if not (0 <= weight <= 1):
            raise ValueError('weight should be between 0 and 1 (inclusive).')

        total_funds = self.total_funds(price)
        value = total_funds * weight
        shares = self.adjust_value(date, price, value, direction)
        return shares

    ####################################################################
    # GET LOGS (trade log, raw trade log)

    def _merge_trades(self, tlog):
        """
        Merge like trades that occur on the same day.
        """

        def _merge(tlog, merge_type):
            """
             Merge entry trades that occur on the same date.
             """
            if merge_type == 'entry':
                tlog['exit_date'] = tlog['exit_date']
            else:
                tlog['entry_date'] = tlog['entry_date']
            tlog['entry_price'] = \
                (tlog['entry_price'] * tlog['qty']).sum() / tlog['qty'].sum()
            tlog['exit_price'] = \
                (tlog['exit_price'] * tlog['qty']).sum() / tlog['qty'].sum()
            tlog['pl_points'] = tlog['pl_points'].sum()
            tlog['pl_cash'] = tlog['pl_cash'].sum()
            tlog['qty'] = tlog['qty'].sum()
            tlog['cumul_total'] = tlog['cumul_total'].tail(1)
            return tlog

        tlog = tlog.groupby('entry_date', group_keys=False) \
                                         .apply(_merge, merge_type='entry') \
                                         .dropna().reset_index(drop=True)
        tlog = tlog.groupby('exit_date', group_keys=False) \
                                        .apply(_merge, merge_type='exit') \
                                        .dropna().reset_index(drop=True)
        return tlog


    def get_log(self, merge_trades=False):
        """
        Return the trade log.

        The trade log consists of the following columns:
        'entry_date', 'entry_price', 'exit_date', 'exit_price',
        'pl_points', 'pl_cash', 'qty', 'cumul_total',
        'direction', 'symbol'.

        Parameters
        ----------
        merge_trade : bool, optional
            True to merge trades that occur on the same date
            (default is False).

        Returns
        -------
        tlog : pd.DataFrame
            The trade log.
        """
        columns = ['entry_date', 'entry_price', 'exit_date', 'exit_price',
                   'pl_points', 'pl_cash', 'qty', 'cumul_total',
                   'direction', 'symbol']
        tlog = pd.DataFrame(self._l, columns=columns)

        if merge_trades:
            tlog = self._merge_trades(tlog)

        return tlog

    def get_log_raw(self):
        """
        Return the raw trade log.

        The trade log consists of the following columns:
        'date', 'seq_num', 'price', 'shares', 'entry_exit',
        'direction', 'symbol'.

        Returns
        -------
        rlog : pd.DataFrame
            The raw trade log.
        """
        columns = ['date', 'seq_num', 'price', 'shares', 'entry_exit', 'direction', 'symbol']
        rlog = pd.DataFrame(self._raw, columns=columns)
        return rlog

########################################################################
# DAILY BALANCE

class TradeState:
    """
    The trade state of OPEN, HOLD, or CLOSE.

    In the Daily Balance log, trade state is given by these
    characters: OPEN='O', HOLD='-', and CLOSE='X'
    """
    OPEN, HOLD, CLOSE = ['O', '-', 'X']

class DailyBal:
    """
    Log for daily balance.
    """

    def __init__(self):
        """
        Initialize instance variables.

        Attributes
        ----------
        _l : list of tuples
            The list of daily balance tuples.
        """
        self._l = []

    def append(self, date, close, high=None, low=None):
        """
        Append a new entry to the daily balance log.

        Parameters
        ----------
        date : str
            The current date.
        close : float
            The balance close value of the day.
        high : float, optional
            The balance high value of the day (default is None,
            which implies that the 'high' is the 'close'.  In other
            words, we are not using intra-day prices).
        low : float, optional
            The balance low value of the day (default is None,
            which implies that the 'low' is the 'close'.  In other
            words, we are not using intra-day prices).

        Returns
        -------
        None
        """
        if high is None:  high = close
        if low  is None:  low  = close

        # calculate daily balance values:
        # date, high, low, close, shares, cash, leverage
        cash = TradeLog.cash
        tlog = list(TradeLog.instance.values())[0]
        shares   = tlog.shares
        high_    = tlog.equity(high)
        low_     = tlog.equity(low)
        close_   = tlog.equity(close)
        leverage = tlog.leverage(close)
        #if (close_ < 0):
        #    print('{} WARNING: Margin Call!!!'
        #          .format(date.strftime('%Y-%m-%d')))

        if tlog.direction == Direction.LONG:
            t = (date, high_, low_, close_, shares, cash, leverage)
        else:
            t = (date, low_, high_, close_, shares, cash, leverage)
        self._l.append(t)

    def get_log(self, tlog):
        """
        Return the daily balance log.

        The daily balance log consists of the following columns:
        'date', 'high', 'low', 'close', 'shares', 'cash', 'leverage'

        Parameters
        ----------
        tlog : pd.DataFrame
            The trade log.

        Returns
        -------
        dbal : pd.DataFrame
            The daily balance log.
        """
        columns = ['date', 'high', 'low', 'close', 'shares', 'cash', 'leverage']
        dbal = pd.DataFrame(self._l, columns=columns)

        def trade_state(row):
            """
            Apply function for adding the `state` column to dbal.

            Convert pandas.timestamp to numpy.datetime64.
            See if there was a entry or exit in tlog on date.
            """
            date = row.date.to_datetime64()
            if date in tlog.entry_date.values:
                state = TradeState.OPEN
            elif date in tlog.exit_date.values:
                state = TradeState.CLOSE
            else:
                state = TradeState.HOLD
            return state

        dbal['state'] = dbal.apply(trade_state, axis=1)
        dbal.set_index('date', inplace=True)
        return dbal
